"""
<description>
Author: Andreas Hellerschmied
"""

import sys
import numpy as np
from scipy.linalg import block_diag

# Imports from other gravtools modules:
from gravtools import settings
from bev_legacy import init


def gauss_markoff(obs_df, stat_df, pol_degree=1):
    """
    Adjustment of gravity meter observations using a Gauss-Markoff model.
    Linear problem => Without linearization - no a priori g values required at non-datum stations.
    Estimated parameters:

     - offset and linear drift per instrument
     - gravity at observed stations

    :return:
    """

    # Options:
    # datum_stations_list = ['0-059-21']
    datum_stations_list = []

    sig0_mugal = 1  # a priori standard deviaiton of unit wheigt of observations [µGal]
    sd_cond_scaling_factor = 1e-1  # Scaling factor for condition "observation2"
    g_sd_defaut_mugal = 10  # Default value for the SD of obsrvations

    # If Sd id not provided in the observation file, use the default value:
    obs_df.loc[obs_df.g_sd_mugal.isna(), 'g_sd_mugal'] = g_sd_defaut_mugal

    stat_df['is_datum'] = False
    obs_df['is_datum'] = False
    if not datum_stations_list:  # is not empty
        # All ÖSGN stations are datum stations
        stat_df['is_datum'] = stat_df.is_oesgn
        obs_df.loc[~obs_df.g_oesgn_mugal.isna(), 'is_datum'] = True
    else:
        # Use datum stations from list:
        for stat in datum_stations_list:
            stat_df.loc[stat_df['punktnummer'] == stat, 'is_datum'] = True
            obs_df.loc[obs_df['punktnummer'] == stat, 'is_datum'] = True
            # Check, if all datum stationsade ÖSGN stations:
            if not stat_df[stat_df['punktnummer'] == stat].is_oesgn.bool():
                print('ERROR: the datum station {} is no ÖSGN station!'.format(stat))

    # obs_df = obs_df.sort_values('is_datum')  # Sort: [obs of new stations; obs of datum stations]
    obs_df = obs_df.reset_index()  # enumeration as index instead of time-tag
    # stat_df = stat_df.sort_values('is_datum')  # Sort: [obs of new stations; obs of datum stations]

    # Lookup-table for station numbers:
    # - e.g. indicating the position of station-related matrix entries
    stat_num_dict = stat_df['punktnummer'].to_dict()
    stat_num_dict = dict((v, k) for k, v in stat_num_dict.items())  # Exchange keys and values

    # Lookup-table for gravimeters:
    gravimeter_type_dict = {}
    gravimeter_type = obs_df.gravimeter_typ.unique()
    for num, grav_type in enumerate(gravimeter_type):
        gravimeter_type_dict[grav_type] = num

    # Get parameters:
    num_gravimeter = obs_df.gravimeter_typ.unique().shape[0]
    num_est_para_per_instr = pol_degree + 1  # number of estimated parameters per gravimeter
    num_est_instr_para = num_est_para_per_instr * num_gravimeter  # total number of estimated instrument parameters
    num_obs = obs_df.shape[0]
    num_stations = stat_df.shape[0]
    num_datum_stations = stat_df[stat_df['is_datum'] == True].shape[0]  # all ÖSGN stations
    num_new_stations = num_stations - num_datum_stations
    num_estimates = num_est_instr_para + num_stations
    degree_of_freedom = num_obs + num_datum_stations - num_estimates  # num_datum_stations = num of conditions

    # Preallocate matices and vectors:
    G = np.zeros([num_obs, num_stations])
    D = np.zeros([num_obs, num_est_instr_para])
    Z = np.zeros([num_datum_stations, num_est_instr_para])
    H = np.zeros([num_datum_stations, num_stations])
    L = np.empty(num_obs)
    L = L.reshape(-1, 1)
    L[:] = np.nan
    h = np.empty(num_datum_stations)
    h = h.reshape(-1, 1)
    h[:] = np.nan

    # Weight matrix:
    P = np.diag((sig0_mugal / obs_df['g_sd_mugal']) ** 2)

    # Set up individual matrices:
    for i_obs, values in obs_df.iterrows():
        # print(values['punktnummer'])
        idx_stat = stat_num_dict[values['punktnummer']]  # Get index of observed station
        G[i_obs, idx_stat] = 1
        L[i_obs] = values['g_red_mugal']
        for grav_type, idx_start in gravimeter_type_dict.items():
            for i_deg in range(0, pol_degree + 1):
                # print(i_deg)
                D[i_obs, idx_start * num_est_para_per_instr + i_deg] = values['dt_h'] ** i_deg  # a_k

    # Conditions:
    i_col = 0
    h_sig_mugal = np.empty(num_datum_stations)
    h_sig_mugal = h_sig_mugal.reshape(-1, 1)
    h_sig_mugal[:] = np.nan
    sd_obs_min_mugal = obs_df['g_sd_mugal'].min()
    for i_stat, values in stat_df.iterrows():
        if values['is_datum']:
            H[i_col, stat_num_dict[values['punktnummer']]] = 1
            h[i_col, 0] = values['g_oesgn_mugal']
            h_sig_mugal[i_col, 0] = sd_obs_min_mugal * sd_cond_scaling_factor
            i_col = i_col + 1

    HP = np.diagflat((sig0_mugal / h_sig_mugal) ** 2)
    P = block_diag(P, HP)
    L = np.concatenate((L, h), axis=0)  # vertical

    A1 = np.concatenate((G, -D), axis=1)  # horizontal
    A2 = np.concatenate((H, Z), axis=1)  # horizontal
    A = np.concatenate((A1, A2), axis=0)  # vertical
    ATP = A.transpose().dot(P)
    N = ATP.dot(A)

    print(' - Defree of Freedom {}'.format(degree_of_freedom))
    print(' - Number of estimates: {}'.format(num_estimates))
    print(' - Rank of N: {}'.format(np.linalg.matrix_rank(N)))
    print(' - Condition of N: {}'.format(np.linalg.cond(N)))
    print(' - Determinate of N: {}'.format(np.linalg.det(N)))
    print(' - Rank of A: {}'.format(np.linalg.matrix_rank(A)))
    print(' - Size of N: {} x {}'.format(N.shape[0], N.shape[0]))
    print(' - Rangdefizit: {}'.format(N.shape[0] - np.linalg.matrix_rank(N)))

    b = ATP.dot(L)
    N_inv = np.linalg.inv(N)
    xd = N_inv.dot(b)  # Ausgeglichene Unbekannte

    v = A.dot(N_inv).dot(ATP).dot(L) - L  # Verbessungeruen, lt. AG1 Equ. (2.40)
    Ld = L + v  # Ausgeglichenen Beobachtungen, lt. AG1 Equ. (2.6)

    # Stochastisches Modell a posteriori:
    Qxdxd = N_inv
    QLdLd = A.dot(N_inv).dot(A.transpose())

    # Empirische Varianz der Gewichtseinheit (a posteriori):
    s02 = (v.transpose().dot(P).dot(v)) / degree_of_freedom  # AG1 (2.99)

    # Kovarianz-Matritzen a posteriori:
    Cxdxd = s02 * Qxdxd
    CLdLd = s02 * QLdLd

    sig_xd = np.sqrt(np.diag(Cxdxd))
    sig_LD = np.sqrt(np.diag(CLdLd))

    # Gewichtsreziproke nach Ansermet AG1 (2.74)
    u = np.trace(P.dot(QLdLd))  # Muss Anzahl der Unbekannten ergeben!
    if not u.round(10) == num_estimates:
        print('WARNING: Gewichtsreziproke nach Ansermet nicht erfüllt (u = {})!'.format(u))

    # Hauptprobe:
    # Funktionaler Zusammenhang zwischen den ausgeglichenen Beobachteungen und den ausgeglichenen Unbekannten genau
    # genug erfüllt?
    w_probe = A.dot(xd) - Ld
    print('Hauptprobe: Max. Widerspuch (Absolutwert) = {}'.format(abs(w_probe).max()))

    print('Empirische Varianz der Gewichtseinheit (a posteriori) = {}'.format(s02.item()))

    # Ausgeglichene Beobachtungen (inkl. Std.Abw. a posteriori):
    print('Ausgeglichene Beobachtungen (Std.Abw.):')
    for i_obs, values in obs_df.iterrows():
        if values.is_datum:
            print(' - {:5.2f}h @ {}: {:12.3f} ({:6.3f}) µGal; v = {:8.3f} µGal *'.format(values['dt_h'],
                                                                                         values['punktnummer'],
                                                                                         Ld[i_obs].item(),
                                                                                         sig_LD[i_obs].item(),
                                                                                         v[i_obs].item()))
        else:
            print(' - {:5.2f}h @ {}: {:12.3f} ({:6.3f}) µGal; v = {:8.3f} µGal'.format(values['dt_h'],
                                                                                       values['punktnummer'],
                                                                                       Ld[i_obs].item(),
                                                                                       sig_LD[i_obs].item(),
                                                                                       v[i_obs].item()))

    # Retrive and print results:
    print('Geschätzte Schwere an Stationen (Std.Abw.):')
    for stat, num in stat_num_dict.items():
        print(' - {}: {:10.3f} µGal ({:6.3f} µGal)'.format(stat, xd[num].item(), sig_xd[num].item()))
    print('Sonstige Parameter (Std.Abw.):')
    for i in range(num_stations, len(xd)):
        print(' - a{}: {:15.3f} ({:6.3f})'.format(i - num_stations, xd[i].item(), sig_xd[i].item()))

    pass


def gauss_markoff_omc(obs_df, stat_df, pol_degree=1):
    """
    Adjustment of gravity meter observations using a Gauss-Markoff model.

    WITH A RRIORI Vales for the estimates
    Estimated parameters:

     - offset and linear drift per instrument
     - gravity at observed stations

    :return:
    """

    # Options:
    # datum_stations_list = ['0-059-21']
    datum_stations_list = []

    sig0_mugal = 1  # a priori standard deviaiton of unit wheigt of observations [µGal]
    sd_cond_scaling_factor = 1e-0  # Scaling factor for conditions ("fictional observations")
    g_sd_defaut_mugal = 10  # Default value for the SD of observations

    # If Sd id not provided in the observation file, use the default value:
    obs_df.loc[obs_df.g_sd_mugal.isna(), 'g_sd_mugal'] = g_sd_defaut_mugal

    stat_df['is_datum'] = False
    obs_df['is_datum'] = False
    if not datum_stations_list:  # is not empty
        # All ÖSGN stations are datum stations
        stat_df['is_datum'] = stat_df.is_oesgn
        obs_df.loc[~obs_df.g_oesgn_mugal.isna(), 'is_datum'] = True
    else:
        # Use datum stations from list:
        for stat in datum_stations_list:
            stat_df.loc[stat_df['punktnummer'] == stat, 'is_datum'] = True
            obs_df.loc[obs_df['punktnummer'] == stat, 'is_datum'] = True
            # Check, if all datum stationsade ÖSGN stations:
            if not stat_df[stat_df['punktnummer'] == stat].is_oesgn.bool():
                print('ERROR: the datum station {} is no ÖSGN station!'.format(stat))

    # obs_df = obs_df.sort_values('is_datum')  # Sort: [obs of new stations; obs of datum stations]
    obs_df = obs_df.reset_index()  # enumeration as index instead of time-tag
    # stat_df = stat_df.sort_values('is_datum')  # Sort: [obs of new stations; obs of datum stations]

    # Lookup-table for station numbers:
    # - e.g. indicating the position of station-related matrix entries
    stat_num_dict = stat_df['punktnummer'].to_dict()
    stat_num_dict = dict((v, k) for k, v in stat_num_dict.items())  # Exchange keys and values

    # Lookup-table for gravimeters:
    gravimeter_type_dict = {}
    gravimeter_type = obs_df.gravimeter_typ.unique()
    for num, grav_type in enumerate(gravimeter_type):
        gravimeter_type_dict[grav_type] = num

    # Get parameters:
    num_gravimeter = obs_df.gravimeter_typ.unique().shape[0]
    num_est_para_per_instr = pol_degree + 1  # number of estimated parameters per gravimeter
    num_est_instr_para = num_est_para_per_instr * num_gravimeter  # total number of estimated instrument parameters
    num_obs = obs_df.shape[0]
    num_stations = stat_df.shape[0]
    num_datum_stations = stat_df[stat_df['is_datum'] == True].shape[0]  # all ÖSGN stations
    num_new_stations = num_stations - num_datum_stations
    num_estimates = num_est_instr_para + num_stations
    degree_of_freedom = num_obs + num_datum_stations - num_estimates  # num_datum_stations = num of conditions

    # Preallocate matices and vectors:
    G = np.zeros([num_obs, num_stations])
    D = np.zeros([num_obs, num_est_instr_para])
    Z = np.zeros([num_datum_stations, num_est_instr_para])
    H = np.zeros([num_datum_stations, num_stations])
    L = np.empty(num_obs)
    L = L.reshape(-1, 1)
    L[:] = np.nan
    h = np.empty(num_datum_stations)
    h = h.reshape(-1, 1)
    h[:] = np.nan

    # ### Model computed observations: ###

    # A priori gravity at observed stations:
    stat_df['g_0_mugal'] = stat_df['g_oesgn_mugal']  # Auf ÖSGN Werte setzen, wenn möglich
    # - Wenn kein apriori g-Wert vorhanden => Auf mittelwert der beobachteten ÖSGN-Punkte setzen
    #   - Besser wäre: Interpoliren, g-Wert vom nächsten Punkt mit VG, etc.
    stat_df.loc[stat_df['g_0_mugal'].isna(), 'g_0_mugal'] = stat_df['g_oesgn_mugal'].mean()
    g0 = stat_df.g_0_mugal.to_numpy()
    g0 = g0.reshape(-1, 1)  # vertical vector

    # A priori coefficients of drit polynomial:
    c0 = np.zeros([pol_degree, 1])

    # a priori gravimeter reading offset a0:
    # - Difference between gravimeter reading and a priori gravity at first observed Datum station
    first_datum_obs = obs_df[obs_df['is_datum'] == True].iloc[0]
    a0 = stat_df[stat_df['punktnummer'] == first_datum_obs.punktnummer].g_oesgn_mugal - first_datum_obs.g_red_mugal
    a0 = a0.to_numpy().reshape(-1, 1)  # vertical vector

    # Vector of a priori parameters X0:
    X0 = np.concatenate((g0, a0, c0), axis=0)  # vertical

    # Weight matrix:
    P = np.diag((sig0_mugal / obs_df['g_sd_mugal']) ** 2)

    # Set up individual matrices:
    for i_obs, values in obs_df.iterrows():
        # print(values['punktnummer'])
        idx_stat = stat_num_dict[values['punktnummer']]  # Get index of observed station
        G[i_obs, idx_stat] = 1
        L[i_obs] = values['g_red_mugal']
        for grav_type, idx_start in gravimeter_type_dict.items():
            for i_deg in range(0, pol_degree + 1):
                # print(i_deg)
                D[i_obs, idx_start * num_est_para_per_instr + i_deg] = values['dt_h'] ** i_deg  # a_k

    # Conditions:
    i_col = 0
    h_sig_mugal = np.empty(num_datum_stations)
    h_sig_mugal = h_sig_mugal.reshape(-1, 1)
    h_sig_mugal[:] = np.nan
    sd_obs_min_mugal = obs_df['g_sd_mugal'].min()
    for i_stat, values in stat_df.iterrows():
        if values['is_datum']:
            H[i_col, stat_num_dict[values['punktnummer']]] = 1
            h[i_col, 0] = values['g_oesgn_mugal']
            # h[i_col, 0] = 0
            h_sig_mugal[i_col, 0] = sd_obs_min_mugal * sd_cond_scaling_factor
            i_col = i_col + 1

    HP = np.diagflat((sig0_mugal / h_sig_mugal) ** 2)
    P = block_diag(P, HP)
    L = np.concatenate((L, h), axis=0)  # vertical

    # Modell-Matrix inkl. Bedingungen (fiktive Beobachtungen) in A2
    A1 = np.concatenate((G, -D), axis=1)  # horizontal
    A2 = np.concatenate((H, Z), axis=1)  # horizontal
    A = np.concatenate((A1, A2), axis=0)  # vertical

    # O-C vector (AG1 (2.31)):
    L0 = A.dot(X0)
    l = L - L0  # AG2: (2.12)

    ATP = A.transpose().dot(P)
    N = ATP.dot(A)

    print(' - Defree of Freedom {}'.format(degree_of_freedom))
    print(' - Number of estimates: {}'.format(num_estimates))
    print(' - Rank of N: {}'.format(np.linalg.matrix_rank(N)))
    print(' - Condition of N: {}'.format(np.linalg.cond(N)))
    print(' - Determinate of N: {}'.format(np.linalg.det(N)))
    print(' - Rank of A: {}'.format(np.linalg.matrix_rank(A)))
    print(' - Size of N: {} x {}'.format(N.shape[0], N.shape[0]))
    print(' - Rangdefizit: {}'.format(N.shape[0] - np.linalg.matrix_rank(N)))

    b = ATP.dot(l)  # AG1 (2.37)
    N_inv = np.linalg.inv(N)
    xd = N_inv.dot(b)  # Ausgeglichene Unbekannte # AG1 (2.37)
    Xd = X0 + xd  # AG1 (2.7)

    v = A.dot(N_inv).dot(ATP).dot(l) - l  # Verbessungeruen, lt. AG1 Equ. (2.40)
    Ld = L + v  # Ausgeglichenen Beobachtungen, lt. AG1 Equ. (2.6)

    # Stochastisches Modell a posteriori:
    Qxdxd = N_inv
    QLdLd = A.dot(N_inv).dot(A.transpose())

    # Empirische Varianz der Gewichtseinheit (a posteriori):
    s02 = (v.transpose().dot(P).dot(v)) / degree_of_freedom  # AG1 (2.99)

    # Kovarianz-Matritzen a posteriori:
    Cxdxd = s02 * Qxdxd
    CLdLd = s02 * QLdLd

    sig_xd = np.sqrt(np.diag(Cxdxd))
    sig_Ld = np.sqrt(np.diag(CLdLd))

    # Gewichtsreziproke nach Ansermet AG1 (2.74)
    u = np.trace(P.dot(QLdLd))  # Muss Anzahl der Unbekannten ergeben!
    if not u.round(10) == num_estimates:
        print('WARNING: Gewichtsreziproke nach Ansermet nicht erfüllt (u = {})!'.format(u))

    # Hauptprobe:
    # Funktionalr Zusammenhang zwischen den ausgeglichenen Beobachteungen und den ausgeglichenen Unbekannten genau
    # genug erfüllt?
    w_probe = A.dot(Xd) - Ld
    print('Hauptprobe: Max. Widerspuch (Absolutwert) = {}'.format(abs(w_probe).max()))

    print('Empirische Varianz der Gewichtseinheit (a posteriori) = {}'.format(s02.item()))

    # Ausgeglichene Beobachtungen (inkl. Std.Abw. a posteriori):
    print('Ausgeglichene Beobachtungen (Std.Abw.):')
    for i_obs, values in obs_df.iterrows():
        if values.is_datum:
            print(' - {:5.2f}h @ {}: {:12.3f} ({:6.3f}) µGal; v = {:8.3f} µGal *'.format(values['dt_h'],
                                                                                         values['punktnummer'],
                                                                                         Ld[i_obs].item(),
                                                                                         sig_Ld[i_obs].item(),
                                                                                         v[i_obs].item()))
        else:
            print(' - {:5.2f}h @ {}: {:12.3f} ({:6.3f}) µGal; v = {:8.3f} µGal'.format(values['dt_h'],
                                                                                       values['punktnummer'],
                                                                                       Ld[i_obs].item(),
                                                                                       sig_Ld[i_obs].item(),
                                                                                       v[i_obs].item()))

    # Retrive and print results:
    print('Geschätzte Schwere an Stationen (Std.Abw.):')
    for stat, num in stat_num_dict.items():
        print(' - {}: {:10.3f} µGal ({:6.3f} µGal)'.format(stat, Xd[num].item(), sig_xd[num].item()))
    print('Sonstige Parameter (Std.Abw.):')
    for i in range(num_stations, len(xd)):
        print(' - a{}: {:15.3f} ({:6.3f})'.format(i - num_stations, Xd[i].item(), sig_xd[i].item()))

    pass


# def vermittelnde_beob_mit_bedingungen(obs_df, stat_df, pol_degree=1):
#     """
#     Adjustment of gravity meter observations using a vermittelnde Beobachtungen mit Bedingungsgleichungen
#     See: AG1, pp.16-17
#     Linear problem => Without linearization - no a priori g values required at non-datum stations.
#     Estimated parameters:
#      - offset and linear drift per instrument
#      - gravity at observed stations
#     :return:
#     """
#     pass


def main(path_oesgn_table, name_oesgn_table, path_obs_file, name_obs_file, out_path):

    obs_df, stat_df, df_oesgn, obs_info_dict = init.read_and_prep_data(path_oesgn_table,
                                                                       name_oesgn_table,
                                                                       path_obs_file,
                                                                       name_obs_file)

    # Gauss-Markoff model
    gauss_markoff(obs_df, stat_df, 1)

    # Gauss - Markoff model with computed observatons and linearization:
    gauss_markoff_omc(obs_df, stat_df, 1)

    # Vermittelnde Beobachtungen mit Bedgingungsgleichungen:
    # adjust_vermittelnde_beob_mit_bedingungen(obs_df, stat_df, 1)

    print('...finished!')


# Run as standalone program:
if __name__ == "__main__":

    if len(sys.argv) == 1:
        print('Eingangsparameter aus options.py bezogen (Default-Parameter).')
        path_oesgn_table = settings.PATH_OESGN_TABLE
        name_oesgn_table = settings.NAME_OESGN_TABLE
        path_obs_file = settings.PATH_OBS_FILE_BEV
        name_obs_file = settings.NAME_OBS_FILE_BEV
        out_path = settings.OUT_PATH

    elif len(sys.argv) == 2:  # Name of obervation file as input argument
        name_obs_file = sys.argv[1]
        # Default-parameters from options-file:
        path_oesgn_table = settings.PATH_OESGN_TABLE
        name_oesgn_table = settings.NAME_OESGN_TABLE
        path_obs_file = settings.PATH_OBS_FILE_BEV
        out_path = settings.OUT_PATH

    else:
        print('Error: Invalid number of input arguments!')
        exit()

    # Run init module:
    main(path_oesgn_table,
         name_oesgn_table,
         path_obs_file,
         name_obs_file,
         out_path)

else:
    # not run as standalone program, but as module
    pass



"""Dialog for loading time series data from TSF files.

Copyright (C) 2023  Andreas Hellerschmied <andreas.hellerschmied@bev.gv.at>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import os

from PyQt5.QtWidgets import QDialog, QFileDialog

from gravtools.gui.dialog_load_tsf_file import Ui_DialogLoadTsfFile


class DialogLoadTsfFile(QDialog, Ui_DialogLoadTsfFile):
    """Dialog for loading time series data from TSF files."""

    def __init__(self, parent=None):
        super().__init__(parent)

        # Run the .setupUi() method to show the GUI
        self.setupUi(self)

        self.pushButton_select_file.pressed.connect(self.select_tsf_file)

    def select_tsf_file(self):
        """Launches dialog to select TSF file and get filename."""
        if self.lineEdit_filename:
            initial_path = self.lineEdit_filename.text()
        else:
            try:
                initial_path = self.parent().parent().campaign.output_directory
            except:
                initial_path = os.getcwd()

        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog
        tsf_filename, _ = QFileDialog.getOpenFileName(self,
                                                      'Select TSF file with campaign data',
                                                      initial_path,
                                                      "TSF file (*.TSF)",
                                                      options=options)
        if not tsf_filename:
            return

        # Valid filename:
        self.lineEdit_filename.setText(tsf_filename)
        try:
            survey_name = os.path.basename(tsf_filename).split('.')[0]
            self.lineEdit_survey_name.setText(survey_name)
        except:
            pass



"""Modelling gravimeters.

Copyright (C) 2024  Andreas Hellerschmied <andreas.hellerschmied@bev.gv.at>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os

import pandas
import pandas as pd
import json
import numpy as np

import gravtools.settings as settings


class Gravimeters:
    """Manages gravity meters.

    Attributes
    ----------
    gravimeters : dict, optional (default = None)
        The keys are tuples of the form (<gravimeter type>, <gravimeter serial number>) and the values are `Gravimeter`
        objects describing the properties of a single gravity meter.

    """
    def __init__(self):
        """Default constructor."""
        self.gravimeters = {}

    def add_gravimeter(self, gravi, verbose=True):
        """Add a gravimeter object and carry out checks.

        Parameters
        ----------
        gravi: `Gravimeter` object
            Gravimeter object that contains all relevant data.
        verbose: bool, optional (default = `True`)
            Print terminal output if `True`.

        Notes
        -----
        Do not overwrite existing data loaded from a gravimeter file by default values and/or values from an observation
        file.
        """
        # Overwrite existing entry only, if the new data was loaded from a gravimeter file:
        if (gravi.gravimeter_type, gravi.serial_number) in self.gravimeters.keys():
            if gravi.data_source_type == 'survey' and self.gravimeters[(gravi.gravimeter_type, gravi.serial_number)].data_source_type == 'file':
                if verbose:
                    print(f'{gravi.name} was already loaded from a gravimeter file => Not added')
                return
        self.gravimeters[(gravi.gravimeter_type, gravi.serial_number)] = gravi
        if verbose:
            print(f'{gravi.name} added.')

    def add_from_json(self, filename_json: str, verbose=True):
        """Load data of one or multiple gravimeters a json file.

        Parameters
        ----------
        filename_json : str
            Name of the json file holding the gravimeter information.
        verbose: bool, optional (default = `True`)
            Print terminal output if `True`.

        Notes
        -----
        If the gravimeter already exists with data loaded from an observation file (= survey) and default scale factors
        (`self.data_source_type` is `survey`), this information is overwritten by the data loaded from the json file.
        """
        if verbose:
            print(f'Load gravimeter data from json file: {filename_json}')
        with open(filename_json, mode='r') as file:
            gravimeters = json.load(file)

        for gm in gravimeters:
            if verbose:
                print(f"Load gravimeter {gm['type'],} with S/N {gm['serial_number']}")
            gravi = Gravimeter(gravimeter_type=gm['type'],
                               manufacturer=gm['manufacturer'],
                               serial_number=gm['serial_number'],
                               height_offset_m=gm['height_offset_m'],
                               data_source=os.path.basename(filename_json),
                               data_source_type='file',
                               description=gm['description'],
                               scale_parameters=gm['calibration'],
                               code=gm['code'],
                               verbose=verbose)
            self.add_gravimeter(gravi)

    def add_from_survey(self, survey, verbose=True):
        """Get the gravimeter data from a single survey object and set unknown attributes to default values.

        Parameters
        ----------
        survey : `gravtools.models.survey.Survey` object
            GravTools survey object.
        verbose: bool, optional (default = `True`)
            Print terminal output if `True`.

        Notes
        -----
        If there is already an entry for a gravimeter loaded from an external source, e.g. from a json file, the
        existing content will NOT be overwritten by default values! In this case `self.data_source_type` is set to
        `file`.

        The default linear scale factor is defined in `settings.DEFAULT_GRAVIMETER_LINEAR_SCALE_FACTOR`
        """
        gravi = Gravimeter(gravimeter_type=survey.gravimeter_type,
                           manufacturer=settings.DEFAULT_GRAVIMETER_MANUFACTURERS[survey.gravimeter_type],
                           serial_number=survey.gravimeter_serial_number,
                           height_offset_m=settings.DEFAULT_GRAVIMETER_REFERENCE_HEIGHT_OFFSET_M[survey.gravimeter_type],
                           data_source=survey.name,
                           data_source_type='survey',
                           description=settings.DEFAULT_GRAVIMETER_DESCRIPTION,
                           scale_parameters=None,  # Use default values!
                           code=settings.DEFAULT_GRAVIMETER_ONE_LETTER_CODES[survey.gravimeter_type],
                           verbose=verbose)
        self.add_gravimeter(gravi)

    def delete_gravimeter(self, gravimeter_type: str, serial_number: str, verbose=True):
        """Delete the gravimeter with the given type and S/N.

        Parameters
        ----------
        gravimeter_type : str
            Gravimeter type.
        serial_number : str
            Instrument serial number
        verbose: bool, optional (default = `True`)
            Print terminal output if `True`.
        """
        del self.gravimeters[(gravimeter_type, serial_number)]
        if verbose:
            print(f'Deleted Gravimeter {gravimeter_type} with S/N {serial_number}.')

    def apply_linear_scale(self, gravimeter_type: str, serial_number: str, gravity_df: pd.DataFrame, verbose=False) -> pd.DataFrame:
        """Apply linear scale factor on observations of the gravimeter with the given type and S/N.

        Parameters
        ----------
        gravimeter_type : str
            Gravimeter type.
        serial_number : str
            Instrument serial number
        gravity_df : pd.DataFrame
            Pandas DataFrame with two columns containing the reference epochs (col. "epoch_dt") as datetime objects and
            the gravity values (col. "g") that are scaled with the linear scaling factor corresponding to the reference
            epochs.
        verbose: bool, optional (default = `False`)
            Print terminal output if `True`.

        Returns
        -------
        pd.DataFrame : "gravity_df" with two additional columns for the scale factors ("linear_scale") and the scaled
        gravity ("g_scaled").
        """
        if verbose:
            print(f'Linear scaling correction for gravimeter {gravimeter_type} ({serial_number})')
        return self.gravimeters[(gravimeter_type, serial_number)].apply_linear_scaling(gravity_df)

    def get_height_offset(self, gravimeter_type: str, serial_number: str):
        """Returns the height offset between the sensor and the top.

        Parameters
        ----------
        gravimeter_type : str
            Gravimeter type.
        serial_number : str
            Instrument serial number
        """
        return self.gravimeters[(gravimeter_type, serial_number)].height_offset_m

    @property
    def number_of_gravimeters(self):
        """Return the number of gravimeters."""
        return len(self.gravimeters)

    def __str__(self):
        return f'{self.number_of_gravimeters} gravimeters with known properties.'

    def __repr__(self):
        return self.__str__()


class Gravimeter:
    """Describes properties of a single gravity meter.

    Attributes
    ----------
    gravimeter_type : str
        Gravimeter type. The type has to be a key in `settings.GRAVIMETER_TYPES`.
    manufacturer : str
        Manufacturer of the instrument.
    serial_number : str
        Instrument serial number.
    height_offset_m : float
        Height offset between the instrument's reference surface (usually instrument top surface) and the sensor level.
    description : str
        Arbitrary description.
    data_source : str
        Source from which the gravimeter data was loaded. This is either the name of a file, or the name of a survey.
    data_source_type : str
        Type of the gravimeter data source. HAs to be listed as key in `settings.GRAVIMETER_DATA_SOURCE_TYPES`.
    scale_df : :py:obj:`pandas.core.frame.DataFrame`
        Pandas Dataframe containing the instrument's calibration parameters valid for time specific time intervals.

        - start_date : :py:obj:`datetime.date`
            Start of the time span at which the scale is valid.
        - end_date : :py:obj:`datetime.date`
            End of the time span at which the scale is valid.
        - linear_factor : float,
            Linear scaling factor valid for the given time span.
        - comment : str
            Arbitrary comment on the scale factor entry, e.g. to document the source or any issues.

    code : str, optional (default = '')
        Short instrument code.

    Notes
    -----
    All epochs have to be given w.r.t. UTC. Dates have bo TZ.
    """

    def __init__(self, gravimeter_type: str, manufacturer: str, serial_number: str, height_offset_m: float,
                 data_source: str, data_source_type: str, description='', scale_parameters=None, code='', verbose=True):
        """Default constructor.

        Parameters
        ----------
        gravimeter_type : str
            Gravimeter type. The type has to be a key in `settings.GRAVIMETER_TYPES`.
        manufacturer: str
            Manufacturer of the instrument.
        serial_number: str
            Instrument serial number.
        height_offset_m: float
            Height offset between the instrument's reference surface (usually instrument top surface) and the sensor
            level.
        data_source : str
            Source from which the gravimeter data was loaded. This is either the name of a file, or the name of a
            survey.
        data_source_type : str
            Type of the gravimeter data source. Has to be listed as key in `settings.GRAVIMETER_DATA_SOURCE_TYPES`.
        description : str, optional (default = '')
            Arbitrary description.
        scale_parameters : list of dicts, optional (default = `None`)
            Each dict contains the start date, the end date and the linear calibration factor. If no data is provided,
            default values are used.
        code: str, optional (default = '')
            Short instrument code.
        verbose: bool, optional (default = `True`)
            Print terminal output if `True`.
        """
        flag_init_scale_factor_zero = False

        # gravimeter_type
        if not isinstance(gravimeter_type, str):
            raise TypeError('"gravimeter_type" has to by a string.')

        # manufacturer
        if not isinstance(manufacturer, str):
            raise TypeError('"manufacturer" has to by a string.')

        # serial_number
        if not isinstance(serial_number, str):
            raise TypeError('"serial_number" has to by a string.')

        # height_offset_m
        if not isinstance(height_offset_m, float):
            raise TypeError('"height_offset_m" has to by a float.')

        # data_source
        if not isinstance(data_source, str):
            raise TypeError('"data_source" has to by a string.')

        # data_source_type
        if not isinstance(data_source_type, str):
            raise TypeError('"data_source_type" has to by a string.')
        if data_source_type not in settings.GRAVIMETER_DATA_SOURCE_TYPES.keys():
            raise RuntimeError(f'"{data_source_type}" not listed settings.GRAVIMETER_DATA_SOURCE_TYPES!')

        # description
        if not isinstance(description, str):
            raise TypeError('"description" has to by a string.')

        # calibrations
        if scale_parameters is None:
            flag_init_scale_factor_zero = True
        else:
            if isinstance(scale_parameters, list):
                if len(scale_parameters) > 0:
                    # Save calibration factors to dataframe
                    scale_dict = {'start_date': [], 'end_date': [], 'linear_factor': [], 'comment': []}
                    for scale in scale_parameters:
                        scale_dict['start_date'].append(scale['start_date'])
                        scale_dict['end_date'].append(scale['end_date'])
                        scale_dict['linear_factor'].append(scale['linear_factor'])
                        scale_dict['comment'].append(scale['comment'])
                        scale_df = pd.DataFrame(scale_dict)
                else:
                    flag_init_scale_factor_zero = True
            else:
                raise TypeError('"code" has to by a string.')
        if flag_init_scale_factor_zero:
            if verbose:
                print(f'Initialize a linear scale factor of 1.0 for gravimeter {manufacturer} '
                      f'{gravimeter_type}')
            scale_dict = {'start_date': [settings.DEFAULT_CALIBRATION_START_DATE],
                          'end_date': [settings.DEFAULT_CALIBRATION_END_DATE],
                          'linear_factor': [settings.DEFAULT_CALIBRATION_LINEAR_FACTOR],
                          'comment': [settings.DEFAULT_CALIBRATION_COMMENT]}
            scale_df = pd.DataFrame(scale_dict)
        scale_df['start_date'] = pd.to_datetime(scale_df['start_date'], utc=True).dt.date
        scale_df['end_date'] = pd.to_datetime(scale_df['end_date'], utc=True).dt.date

        # Check for overlapping time intervals:
        if len(scale_df) > 1:
            for index, row in scale_df.iterrows():
                tmp_filter = (((scale_df['end_date'] <= row['end_date']) & (scale_df['end_date'] >= row['start_date'])) | ((scale_df['start_date'] <= row['end_date']) & (scale_df['start_date'] >= row['start_date'])))
                num_matches = len(tmp_filter[tmp_filter])
                if num_matches > 1:
                    tmp_df = scale_df.loc[tmp_filter]
                    error_str = tmp_df.to_string(index=False)
                    raise RuntimeError(f'Overlapping calibration factor intervals of gravimeter {gravimeter_type} '
                                       f'(S/N {serial_number}):\n' + error_str)

        # code
        if not isinstance(code, str):
            raise TypeError('"code" has to by a string.')

        # Assign and save data. if no errors occurred:
        self.gravimeter_type = gravimeter_type
        self.manufacturer = manufacturer
        self.serial_number = serial_number
        self.height_offset_m = height_offset_m
        self.data_source = data_source
        self.description = description
        self.data_source_type = data_source_type
        self.scale_df = scale_df
        self.code = code

    @property
    def num_scale_factors(self) -> int:
        """Return the number of scale factors."""
        return len(self.scale_df)

    def get_linear_scale_factor(self, epoch: pandas.Timestamp, raise_error_availability: bool = False,
                                verbose=True) -> float:
        """Returns the linear scale factor for the given epoch.

        Parameters
        ----------
        epoch : Pandas.Timestamp (timezone aware with tz=UTC).
            Epoch for which a scale factor should be returned
        raise_error_availability : optional (default = `False`)
            `True` implies that an error is raised if no scale factor is available for the given epoch. Otherwise, a
            scale factor equal `np.nan` is returned without raising an error.
        verbose: bool, optional (default = `True`)
            Print terminal output if `True`.

        Returns
        -------
        float : Scale factor or `numpy.nan` if not available or in case of multiple matches.

        Notes
        -----
        Only the dates are compared when matching epochs!

        """
        tmp_filter = (self.scale_df['start_date'] <= epoch.date()) & (self.scale_df['end_date'] >= epoch.date())
        num_matches = len(tmp_filter[tmp_filter])
        if num_matches == 0:
            if raise_error_availability:
                raise RuntimeError(f'No linear scale factor available for epoch {epoch.isoformat()} and gravimeter {self.name}.')
            return np.nan
        elif num_matches == 1:
            return self.scale_df.loc[tmp_filter, 'linear_factor'].item()
        else:
            if verbose:
                print(f'Warning: {num_matches} matches for epoch {epoch.isoformat()} in the list of scale factors for '
                      f'the gravimeter {self.name}')
            if raise_error_availability:
                raise RuntimeError(f'Warning: {num_matches} matches for epoch {epoch.isoformat()} in the list of scale '
                                   f'factors for the gravimeter {self.name}')
            return np.nan

    def get_linear_scale_factors(self, ref_epochs: list, raise_error_availability: bool = False) -> list[float]:
        """Returns the linear scale factor for the given epoch.

        Parameters
        ----------
        ref_epochs : List of datetime64 objects (timezone aware with tz=UTC).
            Epochs for which a scale factor should be returned
        raise_error_availability : optional (default = `False`)
            `True` implies that an error is raised if no scale factor is available for the given epoch. Otherwise, a
            scale factor equal `np.nan` is returned without raising an error.

        Returns
        -------
        list : Scale factors.

        """
        linear_scale_factors: list[float] = []
        for ref_epoch in ref_epochs:
            linear_scale_factors.append(self.get_linear_scale_factor(ref_epoch, raise_error_availability))
        return linear_scale_factors

    def apply_linear_scaling(self, gravity_df: pd.DataFrame) -> pd.DataFrame:
        """Applies a linear scaling factor on the input gravity data.

        Parameters
        ----------
        gravity_df : pd.DataFrame
            Pandas DataFrame with two columns containing the reference epochs (col. "epoch_dt") as datetime objects and
            the gravity values (col. "g") that are scaled with the linear scaling factor corresponding to the reference
            epochs.

        Notes
        -----
        An error is raised if noe scaling factor is available for the given epoch.

        Returns
        -------
        pd.DataFrame : "gravity_df" with two additional columns for the scale factors ("linear_scale") and the scaled
        gravity ("g_scaled").
        """
        gravity_df['linear_scale'] = self.get_linear_scale_factors(gravity_df['epoch_dt'].to_list(),
                                                                   raise_error_availability=True)
        gravity_df['g_scaled'] = gravity_df['linear_scale'] * gravity_df['g']
        return gravity_df

    @property
    def name(self):
        """Return the name of the gravimeter by combing type and S/N"""
        return self.gravimeter_type + f' ({self.serial_number})'

    def __str__(self):
        return (f"Gravimeter type {self.gravimeter_type} by {self.manufacturer} has {self.num_scale_factors} "
                f"scale factors.")

    def __repr__(self):
        return self.__str__()


if __name__ == '__main__':
    """Main function for testing"""
    filename = '../../data/gravimeters/gravimeters_template.json'
    meters = Gravimeters()
    meters.add_from_json(filename)
    print(meters)

    epochs = pd.date_range(start='2019-06-23', end='2022-08-15', periods=10, tz='UTC').to_list()
    epoch_pd = pd.to_datetime('2023-12-31 12:33:59', utc=True)
    grav = meters.gravimeters[('CG5', '40601')]

    scale = grav.get_linear_scale_factor(epoch_pd)

    scales = grav.get_linear_scale_factors(epochs)

    print('end')



"""Classes for least-squares adjustment of non-differential relative gravimeter observations.

Copyright (C) 2021  Andreas Hellerschmied <andreas.hellerschmied@bev.gv.at>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import numpy as np
import pandas as pd
import datetime as dt
import pytz

from gravtools import settings
from gravtools.models.lsm import LSM, create_hist, global_model_test, tau_test
from gravtools.models import misc
from gravtools import __version__ as GRAVTOOLS_VERSION

# optional imports:
try:
    import geopandas
except ImportError:
    _has_geopandas = False
else:
    _has_geopandas = True


class LSMNonDiff(LSM):
    """Least-squares adjustment of non-differential gravimeter observations with weighted constraints.

    Attributes
    ----------
    setup_obs_df : Pandas DataFrame
        Pandas Dataframes for logging (differential or absolute) setup observations, the related metadata, estimation
        results and statistics. The columns of the dataframe are defined in `self._SETUP_OBS_COLUMNS`.
    observed_stations : list of str
        Unique list of names of observed stations, defining the order of stations (station IDs) for all matrices and
        vectors used in the adjustment scheme.
    """

    _SETUP_OBS_COLUMNS_DTYPES = {
        'survey_name': 'str',
        'ref_epoch_dt': 'datetime64[ns, UTC]',
        'obs_id': 'int',
        'station_name': 'str',
        'setup_id': 'int',
        'g_obs_mugal': 'float',
        'sd_g_obs_mugal': 'float',
        'sd_g_obs_est_mugal': 'float',
        'v_obs_est_mugal': 'float',  # Post-fit residuals
        'sd_v_obs_est_mugal': 'float',  # SD of post-fit residuals
        'w_obs_est_mugal': 'float',
        'r_obs_est': 'float',
        'tau_test_result': 'str',
    }
    _SETUP_OBS_COLUMNS = list(_SETUP_OBS_COLUMNS_DTYPES.keys())

    # Short colum names with max. 10 char suitable for shapefiles:
    _SETUP_OBS_COLUMNS_SHORT = {
        'survey_name': 'survey',
        'ref_epoch_dt': 'epoch_dt',
        'obs_id': 'obs_id',
        'station_name': 'station',
        'setup_id': 'setup',
        'g_obs_mugal': 'g',
        'sd_g_obs_mugal': 'sd_g',
        'sd_g_obs_est_mugal': 'sd_g_est',
        'v_obs_est_mugal': 'v',  # Post-fit residuals
        'sd_v_obs_est_mugal': 'sd_v',  # SD of post-fit residuals
        'w_obs_est_mugal': 'w',
        'r_obs_est': 'r',
        'tau_test_result': 'tau_test',
    }

    # Column names of self.drift_pol_df:
    # - keys: Column names of the pandas dataframe
    # - values: Short description for table headers, etc., in the GUI
    _DRIFT_POL_DF_COLUMNS_DICT = {
        'survey_name': 'Survey',
        'degree': 'Degree',
        'coefficient': 'Coefficient',
        'sd_coeff': 'SD',
        'coeff_unit': 'Unit',
        'ref_epoch_t0_dt': 't0',
    }
    _DRIFT_POL_DF_COLUMNS = list(_DRIFT_POL_DF_COLUMNS_DICT.keys())

    def __init__(self, stat_df, setups, comment='', write_log=True):
        """
        Parameters
        ----------
        stat_df : :py:obj:`gravtools.Station.stat_df`
            The station dataframe contains all relevant station data.
        setups : dict of dicts
            The setups dictionary contains all observation data used for the adjustment. The keys of the dictionary
            are the survey names (str). The items are again keys with the following items:

            - ref_epoch_delta_t_h : datetime object
                Reference epoch for the relative reference times in the column `delta_t_h` in the `setup_df` dataframe.
                The reference epoch is determined as the epoch of the first (active) observation in this survey.
            -  ref_epoch_delta_t_campaign_h : datetime object
                Reference epoch for the relative reference times in the column `delta_t_campaign_h` in the `setup_df`
                dataframe. The reference epoch is determined as the epoch of the first (active) observation in the
                campaign.
            - setup_df : Pandas DataFrame
                Pandas dataframes containing the observation data (see :py:obj:`gravtools.Survey.setup_df`).

        comment : str, optional (default = '')
            Arbitrary comment on the LSM run.
        write_log : bool, optional (default=True)
            Flag that indicates whether log string should be written or not.
        """
        # Call constructor from abstract base class:
        lsm_method = 'LSM_non_diff'
        super().__init__(lsm_method, stat_df=stat_df, setups=setups, comment=comment, write_log=write_log)

    @classmethod
    def from_campaign(cls, campaign, comment='', write_log=True):
        """Constructor that generates and populates the LSM object from a Campaign class object.

        Notes
        -----
        Put checks that are dependent on the LSM method into here! All common checks and preparations are implemented
        in the `from_campaign` method of the parent class `LSM`.

        Parameters
        ----------
        campaign : :py:obj:`gravtools.models.survey.Campaign`
            The campaign object needs to provide setup data for all active surveys and the related station data.
        comment : str, optional (default = '')
            Arbitrary comment on the LSM run.
        write_log : bool, optional (default=True)
            Flag that indicates whether log string should be written or not.

        Returns
        -------
        :py:obj:`.LSMNonDiff`
            Contains all information required for adjusting the campaign.
        """
        return super().from_campaign(campaign, comment=comment, write_log=True)

    def adjust(self, drift_pol_degree=1,
               sig0_mugal=1,
               scaling_factor_datum_observations=1.0,
               add_const_to_sd_of_observations_mugal=0.0,
               scaling_factor_for_sd_of_observations=1.0,
               confidence_level_chi_test=0.95,
               confidence_level_tau_test=0.95,
               drift_ref_epoch_type='survey',
               noise_floor_mugal=0.0,
               verbose=False
               ):  # Confidence level):
        """Run the adjustment based on non-differential observations.

        Parameters
        ----------
        drift_pol_degree : int, optional (default=1)
            Degree of estimated drift polynomial.
        sig0_mugal : int, optional (default=1)
            A priori standard deviation of unit weight of observations [µGal] for the stochastic model of the
            least-squares adjustment.
        scaling_factor_datum_observations : float, optional (default=1.0)
            Factor for scaling the standard deviation (SD) of g of datum stations. The scaled SD is used for
            weighting the direct pseudo observations of g at the datum stations that are introduced as datum
            constraints.
        add_const_to_sd_of_observations_mugal : float, optional (default=0.0)
            The defined additive constant is added to the standard deviation (SD) of setup
            observations in order to scale the SD and the resulting weights to realistic values. In µGal. The scaling
            factor `scaling_factor_for_sd_of_observations` is applied before adding this constant!
        scaling_factor_for_sd_of_observations : float, optional (default=1.0)
            Scaling factor for the standard deviation of the setup observations. `add_const_to_sd_of_observations_mugal`
            is applied after applying the scaling factor!
        confidence_level_chi_test : float, optional (default=0.95)
            Confidence level for the goodness-of-fit test.
        confidence_level_tau_test : float, optional (default=0.95)
            Confidence level for the tau test.
        drift_ref_epoch_type : string ('survey' or 'campaign'), optional (default='survey')
            Defines whether the reference epoch t0 for the estimation of the drift polynomials for each survey in the
            campaign is the reference epoch of the first (active) observation in each survey (option: 'survey') or the
            first (active) observation in the whole campaign (option: 'campaign').
        noise_floor_mugal : float, optional (default=0.0)
            The standard error SE of the estimated gravity values at stations is calculated by SE = sqrt(SD**2 + NF**2),
            where SD is the estimated standard deviation of the station's gravity and NF is the noise floor value
            specified here.
        verbose : bool, optional (default=False)
            If `True`, status messages are printed to the command line, e.g. for debugging and testing
        """

        # Prepare lists and indices:
        # - Observations and parameters:
        self.observed_stations = []
        survey_names = []
        number_of_observations = 0
        setup_ids = []
        for survey_name, setup_data in self.setups.items():
            setup_df = setup_data['setup_df']
            self.observed_stations = self.observed_stations + setup_df['station_name'].to_list()
            number_of_observations = number_of_observations + len(setup_df)
            survey_names.append(survey_name)
            setup_ids = setup_ids + setup_df['setup_id'].to_list()
        self.observed_stations = misc.unique_ordered_list(
            self.observed_stations)  # Unique list of stations => order of stations in matrices!
        number_of_stations = len(self.observed_stations)
        number_of_surveys = len(self.setups)
        # Total number of parameters to be estimated:
        # - 1 g value per station
        # - Drift polynomial coeff. per survey: Polynomial degree * number of surveys
        # - 1 constant instrumental bias per survey
        number_of_parameters = number_of_stations + drift_pol_degree * number_of_surveys + number_of_surveys
        # number_of_diff_obs = number_of_observations - number_of_surveys

        # Check, if setup IDs are unique:
        self.check_unique_setups(setup_ids)

        # - Datum points for weighted constraints:
        # Get dataframe with subset of observed stations only:
        filter_tmp = self.stat_df['station_name'].isin(self.observed_stations)
        self.stat_obs_df = self.stat_df.loc[filter_tmp].copy(deep=True)  # All observed stations
        stat_df_obs_datum = self.stat_obs_df.loc[self.stat_obs_df['is_datum']]
        datum_stations = stat_df_obs_datum['station_name'].to_list()
        number_of_datum_stations = len(datum_stations)
        if number_of_datum_stations < 1:
            raise AssertionError('None of the observed (and active) stations is a datum station '
                                 '(minimum one is required)!')
        if stat_df_obs_datum.loc[:, 'g_mugal'].isnull().any() or stat_df_obs_datum.loc[:, 'sd_g_mugal'].isnull().any():
            raise AssertionError('g and/or sd(g) is missing for at least one datum station! Both values are required'
                                 'for ALL datum stations.')

        if verbose or self.write_log:
            time_now_str = dt.datetime.now(tz=pytz.UTC).strftime('%Y-%m-%d, %H:%M:%S %Z')
            tmp_str = f'#### Adjustment log (non-differential LSM) ####\n'
            tmp_str += f'Processed with GravTools {GRAVTOOLS_VERSION} ({time_now_str})\n'
            tmp_str += f'Comment: {self.comment}\n'
            tmp_str += f'\n'
            tmp_str += f'---- Input data and settings ----\n'
            tmp_str += f'Method: {settings.ADJUSTMENT_METHODS[self.lsm_method]}\n'
            tmp_str += f'Number of surveys: {number_of_surveys}\n'
            tmp_str += f'Number of stations: {number_of_stations}\n'
            tmp_str += f'Number of observations: {number_of_observations}\n'
            tmp_str += f'Number of estimated parameters: {number_of_parameters}\n'
            tmp_str += f'Number of datum stations: {number_of_datum_stations}\n'
            tmp_str += f'Degree of freedom (w/o datum constraints): {number_of_observations - number_of_parameters}\n'
            tmp_str += f'Degree of freedom (with datum constraints): {number_of_observations - number_of_parameters + number_of_datum_stations}\n'
            tmp_str += f'\n'
            tmp_str += f'Degree of drift polynomial: {drift_pol_degree}\n'
            tmp_str += f'One reference epoch for each: {drift_ref_epoch_type}\n'
            tmp_str += f'A priori std. deviation of unit weight [µGal]: {sig0_mugal}\n'
            tmp_str += f'Scaling factor for datum constraints: {scaling_factor_datum_observations}\n'
            tmp_str += f'Scaling factor for SD of setup observations: {scaling_factor_for_sd_of_observations}\n'
            tmp_str += f'Additive const. to SD of setup obs. [µGal]: {add_const_to_sd_of_observations_mugal}\n'
            tmp_str += f'Noise floor for std. error determination [µGal]: {noise_floor_mugal}\n'
            tmp_str += f'Confidence level Chi-test: {confidence_level_chi_test:4.2f}\n'
            tmp_str += f'Confidence level Tau-test: {confidence_level_tau_test:4.2f}\n'
            tmp_str += f'\n'
            tmp_str += f'---- Survey infos ----\n'
            tmp_str += self.survey_info_string
            tmp_str += f'\n'
            tmp_str += f'\n'
            if verbose:
                print(tmp_str)
            if self.write_log:
                self.log_str += tmp_str

        # Initialize matrices:
        if verbose or self.write_log:
            tmp_str = f'---- Set up and populate matrices ----\n'
            if verbose:
                print(tmp_str)
            if self.write_log:
                self.log_str += tmp_str
        # => Initialize complete matrices first and then populate them. This is most efficient!
        # - Observation model:
        mat_A0 = np.zeros([number_of_observations, number_of_parameters])  # Model-matrix
        mat_L0 = np.zeros((number_of_observations, 1))
        mat_sig_ll0 = np.zeros(number_of_observations)
        # - Constraints:
        mat_Ac = np.zeros([number_of_datum_stations, number_of_parameters])  # Model-matrix for constraints
        mat_Lc = np.zeros((number_of_datum_stations, 1))
        #  => Convert to diagonal matrix: P0 = np.diag(np.array([1,2,3,4])) = np.diag(mat_p0)
        mat_sig_llc = np.zeros(number_of_datum_stations)

        # Populate matrices:
        obs_id = -1  # Index of differential observations in vectors mat_L0, rows of mat_A0 and mat_p0
        pd_drift_col_offset = number_of_stations - 1  # Column offset for drift parameters in A-matrix
        survey_count = -1  # Survey counter for indexing the drift parameters in the A-matrix
        g_obs_mugal_list = []
        station_name_list = []
        setup_id_list = []
        sd_g_obs_mugal_list = []
        obs_id_list = []
        survey_names_list = []
        ref_epoch_dt_list = []  # Reference epochs (datetime objects) of observations

        for survey_name, setup_data in self.setups.items():
            setup_df = setup_data['setup_df']
            survey_count += 1

            # Scale and manipulate the SD of setup observations in order to adjust their weights in the adjustment:
            # - Apply scaling factor to SD of setup observations:
            setup_df['sd_g_mugal'] = setup_df['sd_g_mugal'] * scaling_factor_for_sd_of_observations
            # - Add additive constant to SD of setup observations in order to scale them to realistic values:
            setup_df['sd_g_mugal'] = setup_df['sd_g_mugal'] + add_const_to_sd_of_observations_mugal
            if (setup_df['sd_g_mugal'] < 0).any():
                raise AssertionError(
                    f'ERROR: SD of observations ("sd_g_mugal") in survey {survey_name} <= 0 are not allowed! This '
                    f'may be due to the scaling of the SD of observations with additive factors.')

            for index, row in setup_df.iterrows():

                obs_id += 1  # Increment diff. observation ID
                g_obs_mugal = row['g_mugal']
                sd_g_obs_mugal = row['sd_g_mugal']
                station_name = row['station_name']
                setup_id = row['setup_id']
                if drift_ref_epoch_type == 'survey':
                    delta_t_h = row['delta_t_h']  # [hours]
                elif drift_ref_epoch_type == 'campaign':
                    delta_t_h = row['delta_t_campaign_h']  # [hours]

                ref_epoch_dt = row['epoch_dt']

                # Populate matrices and vectors:
                mat_L0[(obs_id, 0)] = g_obs_mugal
                mat_sig_ll0[obs_id] = sd_g_obs_mugal ** 2
                # Partial derivative for g at stations:
                mat_A0[obs_id, self.observed_stations.index(station_name)] = 1
                # Partial derivative for drift polynomial including constant instrumental bias (pol. degree = 0):
                for pd_drift_id in range(drift_pol_degree + 1):
                    mat_A0[obs_id, pd_drift_col_offset + pd_drift_id + 1 + survey_count * (drift_pol_degree + 1)] = \
                        delta_t_h ** pd_drift_id

                # Log data in DataFrame:
                g_obs_mugal_list.append(g_obs_mugal)
                sd_g_obs_mugal_list.append(sd_g_obs_mugal)
                station_name_list.append(station_name)
                obs_id_list.append(obs_id)
                survey_names_list.append(survey_name)
                setup_id_list.append(setup_id)
                ref_epoch_dt_list.append(ref_epoch_dt)

        None_list_placeholder = [None] * len(survey_names_list)
        self.setup_obs_df = pd.DataFrame(list(zip(survey_names_list,
                                                  ref_epoch_dt_list,
                                                  obs_id_list,
                                                  station_name_list,
                                                  setup_id_list,
                                                  g_obs_mugal_list,
                                                  sd_g_obs_mugal_list,
                                                  None_list_placeholder,
                                                  None_list_placeholder,
                                                  None_list_placeholder,
                                                  None_list_placeholder,
                                                  None_list_placeholder,
                                                  None_list_placeholder,
                                                  )),
                                         columns=self._SETUP_OBS_COLUMNS)
        self.setup_obs_df = self.setup_obs_df.astype(self._SETUP_OBS_COLUMNS_DTYPES)

        # - constraints:
        datum_station_id = -1
        for index, row in stat_df_obs_datum.iterrows():
            datum_station_id += 1
            station_name = row['station_name']
            station_id = self.observed_stations.index(station_name)
            mat_Ac[datum_station_id, station_id] = 1  # Partial derivative
            mat_Lc[(datum_station_id, 0)] = row['g_mugal']  # g for datum definition
            sd_mugal_for_weighting = row['sd_g_mugal'] / scaling_factor_datum_observations
            mat_sig_llc[datum_station_id] = sd_mugal_for_weighting ** 2

        # Set up all required matrices:
        mat_A = np.vstack((mat_A0, mat_Ac))  # Eq. (16)
        mat_L = np.vstack((mat_L0, mat_Lc))  # Eq. (16)
        mat_sig_ll = np.diag(np.hstack((mat_sig_ll0, mat_sig_llc)))
        mat_Qll = mat_sig_ll / (sig0_mugal ** 2)
        mat_P = np.linalg.inv(mat_Qll)

        if verbose or self.write_log:
            tmp_str = f'\n'
            tmp_str += f'---- Results and statistics ----\n'
            if verbose:
                print(tmp_str)
            if self.write_log:
                self.log_str += tmp_str

        # Solve equation system:
        mat_N = mat_A.T @ mat_P @ mat_A  # Normal equation matrix
        mat_Qxx = np.linalg.inv(mat_N)  # Co-factor matrix of estimates
        mat_x = mat_Qxx @ (mat_A.T @ mat_P @ mat_L)  # Estimates
        mat_v = (mat_A @ mat_x) - mat_L  # Post-fit residuals
        mat_v = misc.numpy_array_set_zero(mat_v, atol=1e-4)
        mat_Qldld = mat_A @ mat_Qxx @ mat_A.T  # A posteriori Co-factor matrix of adjusted observations
        # mat_Qll = np.linalg.inv(mat_P)
        mat_Qldld = misc.numpy_array_set_zero(mat_Qldld)
        mat_Qvv = mat_Qll - mat_Qldld  # Co-factor matrix of post-fit residuals
        mat_Qvv = misc.numpy_array_set_zero(mat_Qvv)

        # Test: "Gewichtsreziprokenprobe nach Ansermet" (see Skriptum AG1, p. 136, Eq. (6.86))
        u = np.sum(np.diag((mat_P @ mat_Qldld)))  # number of unknown parameters (estimates)
        tmp_diff = np.abs(number_of_parameters - u)
        if np.abs(number_of_parameters - u) > settings.ANSERMET_DIFF_THRESHOLD:
            raise AssertionError(f'"Gewichtsreziprokenprobe nach Ansermet" failed! Difference = {tmp_diff}')
        else:
            if verbose or self.write_log:
                tmp_str = f'# Gewichtsreziprokenprobe nach Ansermet (difference = {tmp_diff}) => Passed!\n'
                if verbose:
                    print(tmp_str)
                if self.write_log:
                    self.log_str += tmp_str

        # Condition of normal equation matrix:
        if verbose or self.write_log:
            cond = np.linalg.cond(mat_N)
            tmp_str = f'Condition of normal equation matrix N = {cond:1.3f}\n'
            if verbose:
                print(tmp_str)
            if self.write_log:
                self.log_str += tmp_str

        # A posteriori variance of unit weight s02:
        dof = mat_A.shape[0] - mat_A.shape[1]  # degree of freedom
        par_r = mat_v.T @ mat_P @ mat_v  # = v^T * P * v
        if dof == 0:
            # s02_a_posteriori_mugal2 = par_r[0][0]
            raise AssertionError('Degree of freedom has to be larger than 0!')
        else:
            s02_a_posteriori_mugal2 = par_r[0][0] / dof  # Eq. (20)

        s0_mugal = np.sqrt(s02_a_posteriori_mugal2)  # A posteriori std. deviation of unit weight
        if verbose or self.write_log:
            tmp_str = f'\n'
            tmp_str += f'A posteriori variance (sd) of unit weight: ' \
                       f'{s02_a_posteriori_mugal2:5.3f} ({s0_mugal:5.3f})\n'
            if verbose:
                print(tmp_str)
            if self.write_log:
                self.log_str += tmp_str

        # ### Statistics and tests ###

        # Convert co-factor matrices to covariance matrices (variances in diagonal vector)
        mat_Cvv = s02_a_posteriori_mugal2 * mat_Qvv  # A posteriori Covariance matrix of post-fit residuals
        mat_Cxx = s02_a_posteriori_mugal2 * mat_Qxx  # A posteriori Covariance matrix of estimated parameters
        mat_Cldld = s02_a_posteriori_mugal2 * mat_Qldld  # A posteriori Covariance matrix of adjusted observations
        # diag_Qxx = np.diag(mat_Qxx)

        # Calculate standard deviations:
        mat_sd_xx = np.sqrt(np.diag(mat_Cxx))  # A posteriori SD of estimates
        mat_sd_ldld = np.sqrt(np.diag(mat_Cldld))  # A posteriori SD of adjusted observations

        # A posteriori SD of residuals:
        # - Check just in case, whether all diagonal elements of the Qvv matrix are positive!
        if (np.diag(mat_Qvv) < 0).any():
            mat_sd_vv = np.sqrt(np.diag(abs(mat_Cvv)))
            if verbose or self.write_log:
                tmp_str = f' - Warning: At least one diagonal element of the Qvv matrix is negative!\n'
                if verbose:
                    print(tmp_str)
                if self.write_log:
                    self.log_str += tmp_str
        else:
            mat_sd_vv = np.sqrt(np.diag(mat_Cvv))

        # creating histogram from residuals
        residual_hist, bin_edges = create_hist(mat_v)  # Calculate histogram

        # goodness-of-fit test
        chi_crit, chi_val, chi_test = global_model_test(confidence_level_chi_test, dof,
                                                        s02_a_posteriori_mugal2, sig0_mugal ** 2)
        if verbose or self.write_log:
            tmp_str = f'\n'
            tmp_str += f'# Goodness-of-fit test results:\n'
            tmp_str += f'# Chi-val  Chi-crt-lower  Chi-crt-upper  Status\n'
            tmp_str += f'  {chi_val:7.3f}  {chi_crit[0]:13.3f}  {chi_crit[1]:13.3f}  {chi_test:s}\n'
            tmp_str += f'\n'
            tmp_str += f'# Histogram of the residuals:\n'
            tmp_str += f'# Lower-edge(µGal)  Upper-edge(µGal)  Frequency\n'
            for loop_1 in range(len(residual_hist)):
                tmp_str += f'  {bin_edges[loop_1]:16.4f}  {bin_edges[loop_1 + 1]:16.4f}  {residual_hist[loop_1]:9.0f}\n'
            if verbose:
                print(tmp_str)
            if self.write_log:
                self.log_str += tmp_str

        # outlier detection effectiveness (redundancy components)
        # diag_Qvv = np.diag(mat_Qvv)
        # mat_R = np.diag(mat_P) * diag_Qvv
        # mat_R is exactly the same as "mat_r"

        # Redundanzanteile (redundancy components):
        # - AG II, pp. 66-71
        mat_r = np.diag(mat_Qvv @ mat_P)

        # Standardized residuals used as test statistics for outlier detection:
        # - AG II, p. 66
        # - Taking care of zero-elements in the mat_sd_vv vector to prevent division by zero errors!
        mat_w = np.zeros(len(mat_v))
        tmp_filter = ~np.isclose(mat_v[:, 0], 0.0)
        mat_w[tmp_filter] = mat_v[tmp_filter, 0] / mat_sd_vv[tmp_filter]

        # Tau test for outlier detection:
        alpha_tau = 1 - confidence_level_tau_test
        tau_test_result, tau_critical_value = tau_test(mat_w=mat_w, dof=dof, alpha=alpha_tau, mat_r=mat_r)
        number_of_outliers = tau_test_result.count("failed")

        tau_test_result_obs = tau_test_result[:number_of_observations]
        tau_test_result_pseudo_obs = tau_test_result[number_of_observations:]

        if verbose or self.write_log:
            tmp_str = f'\n'
            tmp_str += f'# Tau-test results:\n'
            tmp_str += f'Critical value: {tau_critical_value:1.3f}\n'
            tmp_str += f' - Number of detected outliers: {number_of_outliers}\n'
            tmp_str += f' - Number of detected outliers (observations): {tau_test_result_obs.count("failed")}\n'
            tmp_str += f' - Number of detected outliers (datum constraints): {tau_test_result_pseudo_obs.count("failed")}\n'
            tmp_str += f' - Number low redundancy component: {tau_test_result.count("r too small")}\n'
            tmp_str += f'\n'
            if verbose:
                print(tmp_str)
            if self.write_log:
                self.log_str += tmp_str

        # blunder detection parameters calculation
        # sv_tau = 1 - confidence_level_tau_test
        # std_res, tau_val, tau_crt = tau_criterion_test(diag_Qvv, mat_r, mat_v, s02_a_posteriori_mugal2, dof, sv_tau)

        # #### Store results ####
        g_est_mugal = mat_x[0:number_of_stations, 0]
        sd_g_est_mugal = mat_sd_xx[0:number_of_stations]
        drift_pol_coeff = mat_x[number_of_stations:, 0]
        drift_pol_coeff_sd = mat_sd_xx[number_of_stations:]
        sd_g_obs_est_mugal = mat_sd_ldld[:number_of_observations]
        sd_pseudo_obs_est_mugal = mat_sd_ldld[number_of_observations:]
        v_obs_est_mugal = mat_v[:number_of_observations, 0]
        sd_v_obs_est_mugal = mat_sd_vv[:number_of_observations]
        v_pseudo_obs_mugal = mat_v[number_of_observations:, 0]
        # sd_v_pseudo_obs_mugal = mat_sd_vv[number_of_observations:]  # Not used
        w_obs_est_mugal = mat_w[:number_of_observations]
        w_pseudo_obs_mugal = mat_w[number_of_observations:]
        r_obs_est = mat_r[:number_of_observations]
        r_pseudo_obs = mat_r[number_of_observations:]

        # Station related results:
        for idx, stat_name in enumerate(self.observed_stations):
            filter_tmp = self.stat_obs_df['station_name'] == stat_name
            self.stat_obs_df.loc[filter_tmp, 'g_est_mugal'] = g_est_mugal[idx]
            self.stat_obs_df.loc[filter_tmp, 'sd_g_est_mugal'] = sd_g_est_mugal[idx]
            self.stat_obs_df.loc[filter_tmp, 'se_g_est_mugal'] = np.sqrt(
                sd_g_est_mugal[idx] ** 2 + noise_floor_mugal ** 2)
        # Calculate differences to estimates:
        self.stat_obs_df['diff_g_est_mugal'] = self.stat_obs_df['g_est_mugal'] - self.stat_obs_df['g_mugal']
        self.stat_obs_df['diff_se_g_est_mugal'] = self.stat_obs_df['se_g_est_mugal'] - self.stat_obs_df['sd_g_mugal']

        # Drift parameters:
        survey_name_list = []
        degree_list = []
        coefficient_list = []
        sd_coeff_list = []
        coeff_unit_list = []
        ref_epoch_t0_dt_list = []
        tmp_idx = 0
        x_estimate_drift_coeff_names = []
        for survey_name, setup_data in self.setups.items():
            for degree in range(drift_pol_degree + 1):
                survey_name_list.append(survey_name)
                degree_list.append(degree)  # starts with 1
                coefficient_list.append(drift_pol_coeff[tmp_idx])  # * (3600**(degree + 1))  # [µGal/h]
                sd_coeff_list.append(drift_pol_coeff_sd[tmp_idx])
                if drift_ref_epoch_type == 'survey':
                    ref_epoch_t0_dt_list.append(setup_data['ref_epoch_delta_t_h'])
                elif drift_ref_epoch_type == 'campaign':
                    ref_epoch_t0_dt_list.append(setup_data['ref_epoch_delta_t_campaign_h'])
                else:
                    ref_epoch_t0_dt_list.append(None)  # Should not happen!
                if degree == 0:
                    coeff_unit_list.append(f'µGal')
                else:
                    coeff_unit_list.append(f'µGal/h^{degree}')
                tmp_idx += 1
                x_estimate_drift_coeff_names.append(f'{survey_name}-{degree}')
        self.drift_pol_df = pd.DataFrame(list(zip(survey_name_list,
                                                  degree_list,
                                                  coefficient_list,
                                                  sd_coeff_list,
                                                  coeff_unit_list,
                                                  ref_epoch_t0_dt_list)),
                                         columns=self._DRIFT_POL_DF_COLUMNS)

        for idx, v_obs_mugal in enumerate(v_obs_est_mugal):
            filter_tmp = self.setup_obs_df['obs_id'] == idx
            self.setup_obs_df.loc[filter_tmp, 'v_obs_est_mugal'] = v_obs_mugal
            self.setup_obs_df.loc[filter_tmp, 'sd_v_obs_est_mugal'] = sd_v_obs_est_mugal[idx]
            self.setup_obs_df.loc[filter_tmp, 'sd_g_obs_est_mugal'] = sd_g_obs_est_mugal[idx]
            self.setup_obs_df.loc[filter_tmp, 'w_obs_est_mugal'] = w_obs_est_mugal[idx]  # standardized residuals
            self.setup_obs_df.loc[filter_tmp, 'r_obs_est'] = r_obs_est[idx]  # redundancy components
            self.setup_obs_df.loc[filter_tmp, 'tau_test_result'] = tau_test_result_obs[idx]  # str

        # Print results to terminal:
        if verbose or self.write_log:
            tmp_str = f'\n'
            tmp_str += f' - Station data:\n'
            tmp_str += self.stat_obs_df[['station_name', 'is_datum', 'g_mugal', 'g_est_mugal',
                                         'diff_g_est_mugal', 'sd_g_mugal',
                                         'sd_g_est_mugal', 'se_g_est_mugal']].to_string(index=False,
                                                                                        float_format=lambda
                                                                                            x: '{:.1f}'.format(x))
            tmp_str += f'\n\n'
            tmp_str += f' - Drift polynomial coefficients:\n'
            tmp_str += self.drift_pol_df.to_string(index=False, float_format=lambda x: '{:.6f}'.format(x))
            tmp_str += f'\n\n'
            tmp_str += f' - Observations:\n'
            for survey_name in survey_names:
                filter_tmp = self.setup_obs_df['survey_name'] == survey_name
                tmp_str += f'   - Survey: {survey_name}\n'
                tmp_str += self.setup_obs_df.loc[filter_tmp, ['station_name', 'g_obs_mugal',
                                                              'sd_g_obs_mugal', 'sd_g_obs_est_mugal',
                                                              'v_obs_est_mugal']].to_string(index=False,
                                                                                            float_format=lambda
                                                                                                x: '{:.1f}'.format(x))
                tmp_str += f'\n\n'
            tmp_str += f' - Pseudo observations at datum stations (constraints):\n'
            tmp_str += f'Station name  sd [µGal]   v [µGal]   w [µGal]   r [0-1]    Tau test result\n'
            for idx, station_name in enumerate(datum_stations):
                tmp_str += f'{station_name:10}   {sd_pseudo_obs_est_mugal[idx]:8.3}     {v_pseudo_obs_mugal[idx]:+8.3}     {w_pseudo_obs_mugal[idx]:+5.3}     {r_pseudo_obs[idx]:+5.3}  {tau_test_result_pseudo_obs[idx]}\n'
            if verbose:
                print(tmp_str)
            if self.write_log:
                self.log_str += tmp_str

        # Save data/infos to object for later use:
        self.drift_polynomial_degree = drift_pol_degree
        self.sig0_a_priori = sig0_mugal
        self.scaling_factor_datum_observations = scaling_factor_datum_observations
        self.confidence_level_chi_test = confidence_level_chi_test
        self.confidence_level_tau_test = confidence_level_chi_test
        self.number_of_stations = number_of_stations
        self.number_of_datum_stations = number_of_datum_stations
        self.number_of_estimates = number_of_parameters
        self.degree_of_freedom = dof
        self.s02_a_posteriori = s02_a_posteriori_mugal2
        self.Cxx = mat_Cxx
        self.x_estimate_names = self.observed_stations + x_estimate_drift_coeff_names
        self.global_model_test_status = chi_test
        self.number_of_outliers = number_of_outliers
        self.drift_ref_epoch_type = drift_ref_epoch_type

    def export_obs_results_shapefile(self, filename, epsg_code, verbose=True):
        """Export observation-related results from the `setup_obs_df` dataframe to a shapefile.

        Notes
        -----
        This method relies on the optional module `geopandas` (optional dependency).

        Parameters
        ---------
        filename : str
            Name and path of the output shapefile.
        epsg_code: int
            EPSG code of the station coordinates' CRS.
        verbose : bool, optional (default=False)
            `True` implies that status messages are printed to the command line.
        """
        if not _has_geopandas:
            raise ImportError(f'Optional dependency "geopandas" not available, but needed for writing shapefiles!')
        if verbose:
            print(f'Save observation results of lsm run "{self.comment}" to: {filename}')
        setup_obs_df = self.setup_obs_df.copy(deep=True)
        # Change dtypes (just to be saved!):
        setup_obs_df = setup_obs_df.astype(self._SETUP_OBS_COLUMNS_DTYPES)
        setup_obs_df['comment'] = self.comment

        # Get coordinates:
        stat_obs_df_short = self.stat_obs_df[['station_name', 'long_deg', 'lat_deg']].copy(deep=True)
        setup_obs_df = setup_obs_df.merge(stat_obs_df_short, left_on='station_name', right_on='station_name',
                                          how='left')

        # Rename columns to short names with max 10 chars suitable for shapefiles:
        setup_obs_df.rename(columns=self._SETUP_OBS_COLUMNS_SHORT, inplace=True)
        setup_obs_df.drop(columns=['epoch_dt'], inplace=True)  # datetime fields cannot be converted to shapefiles!

        setup_obs_gdf = geopandas.GeoDataFrame(setup_obs_df,
                                               geometry=geopandas.points_from_xy(setup_obs_df['long_deg'],
                                                                                 setup_obs_df['lat_deg']),
                                               crs=epsg_code)
        setup_obs_gdf.to_file(filename)



"""
Adjustment of a drift polynomial of dregreee 1 to 3 using multiple linear regression.
This module is part of the gravtools package.
Author: Andreas Hellerschmied
"""

import pandas as pd
import numpy as np
from sklearn.linear_model import LinearRegression

# Imports from other gravtools modules:
from bev_legacy import utils


def calc_drift_corr_mlr(obs_df, stat_df, pol_degree):
    """
    Berechnung eines Polynoms (Grad n) zur Korrektur der Instrumenten-Drift mittels multipler linearer Regression und
    Korrektur der Gravimeter-Lesungen.
    Berechnung analog zur Fortran Routine 'DRIFT2011.for' von Meurers & Ruess.
    :param obs_df:
    :param stat_df:
    :param pol_degree:
    :return:
    """
    # ##### Multiple Linear Regression: #####

    # Set up obs_df with categorical variables for each point:
    df_short = obs_df.loc[:,
               obs_df.columns.intersection(['punktnummer', 'dt_h', 'g_red_mugal'])]  # Only kep the listed columns

    # Create obs_df with categorical dummy variables for each point name:
    prefix = 'is_pkt'
    df_dummy = pd.get_dummies(df_short, prefix=[prefix])
    categorical_variables = [prefix + '_' + str for str in stat_df.punktnummer]

    # Set up target parameters according to polynomial degree:
    target_parameters = []
    for i in range(1, pol_degree + 1):  # i = 1 to polynomial_degree
        col_name = 'dt_{}'.format(i)
        df_dummy[col_name] = df_dummy['dt_h'].pow(i)
        target_parameters.append(col_name)

    selection_list = target_parameters + categorical_variables

    # Calculate regression:
    mlr = LinearRegression(fit_intercept=False)  # No b0 coefficient estimated
    # mlr = LinearRegression()
    mlr.fit(df_dummy[selection_list], df_dummy['g_red_mugal'])

    pol_coef_dict = {i+1: mlr.coef_[i] for i in range(0, pol_degree)}  # Create dict with polynomial coefficients

    # Store corrected reading (stat_df.g_est_mugal):
    tmp_list = mlr.coef_[-stat_df.shape[0]:]  #
    for i, value in enumerate(tmp_list):
        # print(i, value)
        stat_df.iloc[i, 3] = value

    # Calculate drift correction for all observations:
    poly_coef = utils.prep_polyval_coef(pol_coef_dict)
    np.polyval(poly_coef, obs_df.dt_h)
    obs_df['corr_drift_mugal'] = np.polyval(poly_coef, obs_df.dt_h)

    # Calculate statistics:
    for stat in stat_df.punktnummer:

        # Estimate for this station:
        g_est_mugal = stat_df.loc[stat_df['punktnummer'] == stat].g_est_mugal.values[0]  # Schätzwert an der Station

        # Calculate difference between drift corrected reading and the estimates gravity reading at this point (abw):
        obs_df.loc[obs_df['punktnummer'] == stat, 'abw_mugal'] = \
            obs_df.loc[obs_df['punktnummer'] == stat, 'g_red_mugal'] - \
            obs_df.loc[obs_df['punktnummer'] == stat, 'corr_drift_mugal'] - g_est_mugal

        # Mind. 2 Beobachtungen (= Driftpunkt)?
        if stat_df.loc[stat_df['punktnummer'] == stat].is_drift_point.bool():
            num_of_obs = obs_df.loc[obs_df['punktnummer'] == stat].shape[0]
            # Warum "+4"? Eher willkürlich in DRIFT2011 gesetzt, oder? 4 = 2*2.
            stat_df.loc[stat_df['punktnummer'] == stat, 'sig_g_est_mugal'] = \
                np.sqrt((sum(obs_df.loc[obs_df['punktnummer'] == stat, 'abw_mugal'] ** 2) / (num_of_obs - 1)) + 4)

    pol_coef_sig_mugal = np.sqrt(
        sum(obs_df.abw_mugal ** 2) / (obs_df.shape[0] - 1 - sum(stat_df.is_drift_point == False)))

    # Sigma for all estimats with only one observation:
    stat_df.loc[~stat_df['is_drift_point'], 'sig_g_est_mugal'] = np.sqrt(pol_coef_sig_mugal ** 2 + 25)

    results_dict = dict()
    results_dict['pol_coef'] = pol_coef_dict
    results_dict['pol_coef_sig_mugal'] = pol_coef_sig_mugal
    results_dict['stat_df'] = stat_df
    results_dict['obs_df'] = obs_df  # Erweitert mit Verbessungen und Drift-Korekturen je Beobachtung

    return results_dict



"""Model classes for PyQt's model view architecture.

Copyright (C) 2024  Andreas Hellerschmied <andreas.hellerschmied@bev.gv.at>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from PyQt5.QtCore import QAbstractTableModel, Qt
from PyQt5.QtWidgets import QMessageBox
import datetime as dt


NONE_REPRESENTATION_IN_TABLE_VIEW = ''  # Representation of None values in table views in the GUI


class GravimeterScaleFactorTableModel(QAbstractTableModel):
    """Model for displaying the scale factor table of a gravimeter."""

    _DECIMAL_PLACES_PER_FLOAT_COLUMN = {
        'linear_factor': 6,
    }

    # Column names (keys) and short description (values):
    _PLOT_COLUMNS_DICT = {
        'start_date': 'Start date',
        'end_date': 'End date',
        'linear_factor': 'Linear scale factor',
        'comment': 'Comment',
    }
    _PLOT_COLUMNS = list(_PLOT_COLUMNS_DICT.keys())  # Actual list of columns to be shown

    def __init__(self, gravimeters):
        """Initialize the gravimeter scale factor table view model.

        Parameters
        ----------
        gravimeters : py.obj:`gravtools.models.gravimeters.Gravimeters` objects
        """
        QAbstractTableModel.__init__(self)
        self._gravimeters = {}
        self._data = None
        self.load_gravimeters(gravimeters)
        self._gravimeter_index = None  # Tuple: (type, S/N)
        self._data_column_names = None

    def load_gravimeters(self, gravimeters):
        """Load gravimeters object.

        Notes
        -----
        The data is assigned by reference.
        """
        self._gravimeters = gravimeters

    def update_view_model(self, gravimeter_type: str, serial_number: str):
        """Update the `_data` DataFrame that hold the actual data that is displayed."""
        flag_error_init = False
        try:
            # results_drift_df = self._lsm_runs[lsm_run_index].get_results_drift_df
            scale_df = self._gravimeters.gravimeters[(gravimeter_type, serial_number)].scale_df
        except KeyError:
            QMessageBox.critical(self.parent(), 'Error!', f'Gravimeter {gravimeter_type} (S/N: {serial_number}) not found!')
        else:
            if scale_df is None:  # E.g. no results in campaign yet
                flag_error_init = True
            else:
                self._data = scale_df
                self._gravimeter_index = (gravimeter_type, serial_number)
                self._data_column_names = self._data.columns.to_list()
        if flag_error_init:
            self._data = None
            self._gravimeters = None
            self._gravimeter_index = None
            self._data_column_names = None

    def rowCount(self, parent=None):
        if self._data is not None:
            return self._data.shape[0]
        else:
            return 0

    def columnCount(self, parent=None):
        if self._data is not None:
            return self._data.shape[1]
        else:
            return 0

    def data(self, index, role=Qt.DisplayRole):
        if index.isValid():
            if role == Qt.DisplayRole or role == Qt.UserRole:
                value = self._data.iloc[index.row(), index.column()]
                column_name = self._data_column_names[index.column()]
                # Custom formatter (string is expected as return type):
                if value is None:  #
                    return NONE_REPRESENTATION_IN_TABLE_VIEW
                elif isinstance(value, float):
                    if value != value:  # True, if value is "NaN"
                        return NONE_REPRESENTATION_IN_TABLE_VIEW
                    else:
                        if column_name in self._DECIMAL_PLACES_PER_FLOAT_COLUMN.keys():
                            num_dec_places = self._DECIMAL_PLACES_PER_FLOAT_COLUMN[column_name]
                            return '{1:.{0}f}'.format(num_dec_places, value)
                        else:
                            return str(value)
                elif isinstance(value, dt.datetime):
                    return value.strftime("%Y-%m-%d, %H:%M:%S")
                elif isinstance(value, dt.date):
                    return value.strftime("%Y-%m-%d")
                else:  # all other
                    return str(value)

            if role == Qt.TextAlignmentRole:
                value = self._data.iloc[index.row(), index.column()]
                if isinstance(value, int) or isinstance(value, float):
                    # Align right, vertical middle.
                    return Qt.AlignVCenter + Qt.AlignRight
        return None

    def headerData(self, section, orientation, role):
        # section is the index of the column/row.
        if role == Qt.DisplayRole:
            if self._data is not None:
                if orientation == Qt.Horizontal:
                    return self._PLOT_COLUMNS_DICT[str(self._data.columns[section])]
                if orientation == Qt.Vertical:
                    return str(self._data.index[section])

    def get_short_column_description(self, column_name: str) -> str:
        """Returns the short description of the model column."""
        try:
            return self._PLOT_COLUMNS_DICT[column_name]
        except AttributeError:
            return ''



"""Classes for the estimation of vertical gravity gradient by least-squares adjustment of differential relative
gravimeter observations.

Copyright (C) 2021  Andreas Hellerschmied <andreas.hellerschmied@bev.gv.at>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import numpy as np
import pandas as pd
import datetime as dt
import pytz
# from matplotlib import pyplot as plt

from gravtools import settings
from gravtools.models.lsm import LSM, create_hist, global_model_test, tau_test
from gravtools.models import misc
from gravtools import __version__ as GRAVTOOLS_VERSION


class VGLSM(LSM):
    """VG estimation by least-squares adjustment of non-differential gravimeter observations.

    Notes
    -----
    The estimation of vertical gradients is restricted to th observations of a single survey and one station!


    Attributes
    ----------
    setup_obs_df : Pandas DataFrame
        Pandas Dataframes for logging (differential or absolute) setup observations, the related metadata, estimation
        results and statistics. The columns of the dataframe are defined in `self._SETUP_OBS_COLUMNS`.
    observed_stations : list of str
        Unique list of names of observed stations, defining the order of stations (station IDs) for all matrices and
        vectors used in the adjustment scheme.
    """

    # Column names of self.setup_obs_df:
    # - keys: Column names of the pandas dataframe
    # - values: Short description for table headers, etc., in the GUI
    _SETUP_OBS_COLUMNS_DICT = {
        'survey_name': 'Survey',
        'ref_epoch_dt': 'Epoch',
        'obs_id': 'Obs. ID',
        'station_name': 'Station',
        'setup_id': 'Setup ID',
        'g_obs_mugal': 'g [µGal]',
        'sd_g_obs_mugal': 'SD [µGal]',
        'dhf_sensor_m': 'Sensor height [m]',
        'sd_g_obs_est_mugal': 'SD_est [µGal]',
        'v_obs_est_mugal': 'Residuals [µGal]',  # Post fit residuals
        'sd_v_obs_est_mugal': 'SD_v [µGal]',  # SD of post-fit residuals
        'w_obs_est_mugal': 'Std. Residual []',
        'r_obs_est': 'Redundancy []',
        'tau_test_result': 'Outlier Test',
    }
    _SETUP_OBS_COLUMNS = list(_SETUP_OBS_COLUMNS_DICT.keys())

    # Column names of self.drift_pol_df:
    # - keys: Column names of the pandas dataframe
    # - values: Short description for table headers, etc., in the GUI
    _DRIFT_POL_DF_COLUMNS_DICT = {
        'survey_name': 'Survey',
        'degree': 'Degree',
        'coefficient': 'Coefficient',
        'sd_coeff': 'SD',
        'coeff_unit': 'Unit',
        'ref_epoch_t0_dt': 't0',
    }
    _DRIFT_POL_DF_COLUMNS = list(_DRIFT_POL_DF_COLUMNS_DICT.keys())

    # Column names of self.vg_pol_df:
    # - keys: Column names of the pandas dataframe
    # - values: Short description for table headers, etc., in the GUI
    _VG_POL_DF_COLUMNS_DICT = {
        'degree': 'Degree',
        'coefficient': 'Coefficient',
        'sd_coeff': 'SD',
        'coeff_unit': 'Unit',
    }
    _VG_POL_DF_COLUMNS = list(_VG_POL_DF_COLUMNS_DICT.keys())

    def __init__(self, stat_df, setups, comment='', write_log=True):
        """
        Parameters
        ----------
        stat_df : :py:obj:`gravtools.Station.stat_df`
            The station dataframe contains all relevant station data.
        setups : dict of dicts
            The setups dictionary contains all observation data used for the adjustment. The keys of the dictionary
            are the survey names (str). The items are again keys with the following items:

            - ref_epoch_delta_t_h : datetime object
                Reference epoch for the relative reference times in the column `delta_t_h` in the `setup_df` dataframe.
                The reference epoch is determined as the epoch of the first (active) observation in this survey.
            -  ref_epoch_delta_t_campaign_h : datetime object
                Reference epoch for the relative reference times in the column `delta_t_campaign_h` in the `setup_df`
                dataframe. The reference epoch is determined as the epoch of the first (active) observation in the campaign.
            - setup_df : Pandas DataFrame
                Pandas dataframes containing the observation data (see :py:obj:`gravtools.Survey.setup_df`).

        comment : str, optional (default = '')
            Arbitrary comment on the LSM run.
        write_log : bool, optional (default=True)
            Flag that indicates whether log string should be written or not.
        """
        # Call constructor from abstract base class:
        lsm_method = 'VG_LSM_nondiff'
        super().__init__(lsm_method, stat_df=stat_df, setups=setups, comment=comment, write_log=write_log)

    @classmethod
    def from_campaign(cls, campaign, comment='', write_log=True):
        """Constructor that generates and populates the LSM object from a Campaign class object.

        Notes
        -----
        Put checks that are dependent on the LSM method into here! All common checks and preparations are implemented
        in the `from_campaign` method of the parent class `LSM`.

        Parameters
        ----------
        campaign : :py:obj:`gravtools.models.survey.Campaign`
            The campaign object needs to provide setup data for all active surveys and the related station data.
        comment : str, optional (default = '')
            Arbitrary comment on the LSM run.
        write_log : bool, optional (default=True)
            Flag that indicates whether log string should be written or not.

        Returns
        -------
        :py:obj:`.VGLSM`
            Contains all information required for adjusting the campaign.
        """
        # Station data:
        # - Just one station (at different height levels) allowed => Check coordinates!
        if campaign.stations is None:
            raise AssertionError(f'The campaign "{campaign.campaign_name}" does not contain any station data!')
        else:
            if not hasattr(campaign.stations, 'stat_df'):
                raise AssertionError(f'The campaign "{campaign.campaign_name}" has not station dataframe!')
            else:
                if len(campaign.stations.stat_df) == 0:
                    raise AssertionError(f'The campaign "{campaign.campaign_name}" has an empty station dataframe!')
                else:  # Just one station allowed => Check the coordinates (lon, lat)!
                    stat_df_observed = campaign.stations.stat_df.loc[campaign.stations.stat_df.is_observed]
                    if (len(stat_df_observed['lat_deg'].unique()) > 1) or (
                            len(stat_df_observed['long_deg'].unique()) > 1):
                        raise AssertionError(f'This campaign contains observed stations with differing latitudes and/or'
                                             f' longitudes. This is an indicator for observations of more than one '
                                             f'stations which is not allowed.')

        # Survey data:
        # - Just one survey allowed!
        if campaign.surveys is None:
            raise AssertionError(f'The campaign "{campaign.campaign_name}" does not contain any survey data!')
        else:
            if len(campaign.surveys) > 1:
                raise AssertionError(
                    f'The campaign "{campaign.campaign_name}" contains more than one survey! The VG estimation is '
                    f'restricted to just one survey.')

        # Check if the reduced setup observations refer to the sensor height of the instrument:
        # Loop over surveys in campaign (just one...):
        for survey_name, survey in campaign.surveys.items():
            if survey.keep_survey and survey.is_active:
                if survey.red_reference_height_type != 'sensor_height':
                    raise AssertionError(f'In survey {survey_name} the reduced gravity values do not refer to the '
                                         f'sensor height! Change the reference height to "Sensor" and recalculate the '
                                         f'setup observation data.')

        return super().from_campaign(campaign, comment=comment, write_log=True)

    def adjust(self, drift_pol_degree=1,
               vg_polynomial_degree=1,
               vg_polynomial_ref_height_offset_m=0.0,
               sig0_mugal=1,
               confidence_level_chi_test=0.95,
               confidence_level_tau_test=0.95,
               verbose=False
               ):
        """Run the VG adjustment based on non-differential observations.

        Parameters
        ----------
        drift_pol_degree : int, optional (default=1)
            Degree of estimated drift polynomial.
        vg_polynomial_degree : int, optional (default=1)
            Degree of the estimated polynomial of the vertical gravity gradient. Valid values are 1, 2 or 3.
        vg_polynomial_ref_height_offset_m : float, optional (default=0.0)
            Vertical offset [m] between the reference height of the control point and the zero-level of the estimated
            VG polynomial. A positive offset describes an offset above the control point.
        sig0_mugal : int, optional (default=1)
            A priori standard deviation of unit weight of observations [µGal] for the stochastic model of the
            least-squares adjustment.
        confidence_level_chi_test : float, optional (default=0.95)
            Confidence level for the goodness-of-fit test.
        confidence_level_tau_test : float, optional (default=0.95)
            Confidence level for the tau test.
        verbose : bool, optional (default=False)
            If True, status messages are printed to the command line, e.g. for debugging and testing

        Notes
        -----
        """

        # Initial checks:
        if (vg_polynomial_degree < 1) or (vg_polynomial_degree > 3):
            raise AssertionError(f'Invalid degree of the VG polynomial (degree = {vg_polynomial_degree})! Valid values '
                                 f'are 1 to 3.')

        # Prepare lists and indices:
        # - Observations and parameters:
        self.observed_stations = []
        survey_names = []
        number_of_observations = 0
        setup_ids = []

        for survey_name, setup_data in self.setups.items():
            setup_df = setup_data['setup_df']
            self.observed_stations = self.observed_stations + setup_df['station_name'].to_list()
            number_of_observations = number_of_observations + len(setup_df)
            survey_names.append(survey_name)
            setup_ids = setup_ids + setup_df['setup_id'].to_list()
        self.observed_stations = misc.unique_ordered_list(
            self.observed_stations)  # Unique list of stations => order of stations in matrices!
        number_of_stations = len(self.observed_stations)
        number_of_surveys = len(self.setups)  # Has to be 1 anyway

        # ### Station data ###
        # Get dataframe with subset of observed stations only:
        filter_tmp = self.stat_df['station_name'].isin(self.observed_stations)
        self.stat_obs_df = self.stat_df.loc[filter_tmp].copy(deep=True)  # All observed stations
        # Drop columns that are not required at VG estimation:
        self.stat_obs_df.drop(columns=['g_mugal', 'sd_g_mugal', 'is_datum'], inplace=True)
        # Calculate statistical data for the sensor height (relevant information for VG estimation):
        tmp_df = setup_df.loc[:, ['station_name', 'dhf_sensor_m']].groupby('station_name').mean().rename(
            columns={"dhf_sensor_m": "dhf_sensor_mean_m"})
        self.stat_obs_df = self.stat_obs_df.merge(tmp_df, left_on='station_name', right_on='station_name', how='left')
        tmp_df = setup_df.loc[:, ['station_name', 'dhf_sensor_m']].groupby('station_name').std().rename(
            columns={"dhf_sensor_m": "dhf_sensor_std_m"})
        self.stat_obs_df = self.stat_obs_df.merge(tmp_df, left_on='station_name', right_on='station_name', how='left')
        tmp_df = setup_df.loc[:, ['station_name', 'dhf_sensor_m']].groupby('station_name').min().rename(
            columns={"dhf_sensor_m": "dhf_sensor_min_m"})
        self.stat_obs_df = self.stat_obs_df.merge(tmp_df, left_on='station_name', right_on='station_name', how='left')
        tmp_df = setup_df.loc[:, ['station_name', 'dhf_sensor_m']].groupby('station_name').max().rename(
            columns={"dhf_sensor_m": "dhf_sensor_max_m"})
        self.stat_obs_df = self.stat_obs_df.merge(tmp_df, left_on='station_name', right_on='station_name', how='left')

        if number_of_surveys > 1:
            raise AssertionError(
                f'Invalid number of surveys ({number_of_surveys})! Only one survey allowed for VG estimation.')
        # Total number of parameters to be estimated:
        # - Drift polynomial coeff.: Polynomial degree * number of surveys
        # - 1 constant instrumental bias per survey
        # - VG polynomial coeff.: Polynomial degree * number of surveys
        number_of_parameters = drift_pol_degree + vg_polynomial_degree + 1

        # Check, if setup IDs are unique:
        self.check_unique_setups(setup_ids)

        if verbose or self.write_log:
            time_now_str = dt.datetime.now(tz=pytz.UTC).strftime('%Y-%m-%d, %H:%M:%S %Z')
            tmp_str = f'#### Adjustment log (VG LGM estimation based on non-differential observations) ####\n'
            tmp_str += f'Processed with GravTools {GRAVTOOLS_VERSION} ({time_now_str})\n'
            tmp_str += f'Comment: {self.comment}\n'
            tmp_str += f'\n'
            tmp_str += f'---- Input data and settings ----\n'
            tmp_str += f'Method: {settings.ADJUSTMENT_METHODS[self.lsm_method]}\n'
            tmp_str += f'Number of stations (i.e. height levels): {number_of_stations}\n'
            tmp_str += f'Number of observations: {number_of_observations}\n'
            tmp_str += f'Number of estimated parameters: {number_of_parameters}\n'
            tmp_str += f'Degree of freedom: {number_of_observations - number_of_parameters}\n'
            tmp_str += f'\n'
            tmp_str += f'Degree of drift polynomial: {drift_pol_degree}\n'
            tmp_str += f'Degree of VG polynomial: {vg_polynomial_degree}\n'
            tmp_str += f'VG polynomial height offset [m]: {vg_polynomial_ref_height_offset_m:4.3f}\n'
            tmp_str += f'A priori std. deviation of unit weight [µGal]: {sig0_mugal}\n'
            tmp_str += f'Confidence level Chi-test: {confidence_level_chi_test:4.2f}\n'
            tmp_str += f'Confidence level Tau-test: {confidence_level_tau_test:4.2f}\n'
            tmp_str += f'\n'
            tmp_str += f'---- Survey infos ----\n'
            tmp_str += self.survey_info_string
            tmp_str += f'\n'
            tmp_str += f'\n'
            if verbose:
                print(tmp_str)
            if self.write_log:
                self.log_str += tmp_str

        # Initialize matrices:
        if verbose or self.write_log:
            tmp_str = f'---- Set up and populate matrices ----\n'
            if verbose:
                print(tmp_str)
            if self.write_log:
                self.log_str += tmp_str
        # => Initialize complete matrices first and then populate them. This is most efficient!
        # - Observation model:
        mat_A = np.zeros([number_of_observations, number_of_parameters])  # Design matrix
        mat_L = np.zeros((number_of_observations, 1))  # Observations
        mat_sig_ll = np.zeros(number_of_observations)  # Variances of obs.

        # Populate matrices:
        obs_id = -1  # Index of differential observations in vectors mat_L, rows of mat_A and mat_p0
        survey_count = -1  # Survey counter for indexing the drift parameters in the A-matrix
        g_obs_mugal_list = []
        station_name_list = []
        setup_id_list = []
        sd_g_obs_mugal_list = []
        obs_id_list = []
        survey_names_list = []
        ref_epoch_dt_list = []  # Reference epochs (datetime objects) of observations
        dhf_sensor_m_list = []  # Height above the control point for VG estimation

        for survey_name, setup_data in self.setups.items():
            setup_df = setup_data['setup_df']
            survey_count += 1

            # Add offset to the height of the sensor above the control point:
            # setup_df['dhf_sensor_m'] = setup_df['dhf_sensor_m'] + vg_polynomial_ref_height_offset_m
            if (setup_df['sd_g_mugal'] < 0).any():
                raise AssertionError(
                    f'ERROR: SD of observations ("sd_g_mugal") in survey {survey_name} <= 0 are not allowed!')

            for index, row in setup_df.iterrows():
                obs_id += 1  # Increment diff. observation ID
                g_obs_mugal = row['g_mugal']
                sd_g_obs_mugal = row['sd_g_mugal']
                station_name = row['station_name']
                setup_id = row['setup_id']
                delta_t_h = row['delta_t_h']  # [hours]
                ref_epoch_dt = row['epoch_dt']
                dhf_sensor_m = row['dhf_sensor_m'] + vg_polynomial_ref_height_offset_m

                # Populate matrices and vectors:
                mat_L[(obs_id, 0)] = g_obs_mugal
                mat_sig_ll[obs_id] = sd_g_obs_mugal ** 2
                # Partial derivative for drift polynomial including constant instrumental bias (pol. degree = 0):
                for pd_drift_id in range(drift_pol_degree + 1):
                    mat_A[obs_id, pd_drift_id] = \
                        delta_t_h ** pd_drift_id
                # Partial derivative for VG polynomial:
                for pd_vg_id in range(vg_polynomial_degree):
                    mat_A[obs_id, drift_pol_degree + 1 + pd_vg_id] = \
                        dhf_sensor_m ** (pd_vg_id + 1)

                # Log data in DataFrame:
                g_obs_mugal_list.append(g_obs_mugal)
                sd_g_obs_mugal_list.append(sd_g_obs_mugal)
                station_name_list.append(station_name)
                obs_id_list.append(obs_id)
                survey_names_list.append(survey_name)
                setup_id_list.append(setup_id)
                ref_epoch_dt_list.append(ref_epoch_dt)
                dhf_sensor_m_list.append(dhf_sensor_m)

        None_list_placeholder = [None] * len(survey_names_list)
        self.setup_obs_df = pd.DataFrame(list(zip(survey_names_list,
                                                  ref_epoch_dt_list,
                                                  obs_id_list,
                                                  station_name_list,
                                                  setup_id_list,
                                                  g_obs_mugal_list,
                                                  sd_g_obs_mugal_list,
                                                  dhf_sensor_m_list,
                                                  None_list_placeholder,
                                                  None_list_placeholder,
                                                  None_list_placeholder,
                                                  None_list_placeholder,
                                                  None_list_placeholder,
                                                  None_list_placeholder,
                                                  )),
                                         columns=self._SETUP_OBS_COLUMNS)

        # Set up all required matrices:
        mat_sig_ll = np.diag(mat_sig_ll)
        mat_Qll = mat_sig_ll / (sig0_mugal ** 2)
        mat_P = np.linalg.inv(mat_Qll)

        if verbose or self.write_log:
            tmp_str = f'\n'
            tmp_str += f'---- Results and statistics ----\n'
            if verbose:
                print(tmp_str)
            if self.write_log:
                self.log_str += tmp_str

        # Solve equation system:
        mat_N = mat_A.T @ mat_P @ mat_A  # Normal equation matrix
        mat_Qxx = np.linalg.inv(mat_N)  # Co-factor matrix of estimates
        mat_x = mat_Qxx @ (mat_A.T @ mat_P @ mat_L)  # Estimates
        mat_v = (mat_A @ mat_x) - mat_L  # Post-fit residuals
        mat_v = misc.numpy_array_set_zero(mat_v, atol=1e-4)  # ToDo: Check atol!
        mat_Qldld = mat_A @ mat_Qxx @ mat_A.T  # A posteriori Co-factor matrix of adjusted observations
        mat_Qldld = misc.numpy_array_set_zero(mat_Qldld)
        # mat_Qll = np.linalg.inv(mat_P)
        mat_Qvv = mat_Qll - mat_Qldld  # Co-factor matrix of post-fit residuals
        mat_Qvv = misc.numpy_array_set_zero(mat_Qvv)

        # Test: "Gewichtsreziprokenprobe nach Ansermet" (see Skriptum AG1, p. 136, Eq. (6.86))
        u = np.sum(np.diag((mat_P @ mat_Qldld)))  # number of unknown parameters (estimates)
        tmp_diff = np.abs(number_of_parameters - u)
        if np.abs(number_of_parameters - u) > settings.ANSERMET_DIFF_THRESHOLD:
            raise AssertionError(f'"Gewichtsreziprokenprobe nach Ansermet" failed! Difference = {tmp_diff}')
        else:
            if verbose or self.write_log:
                tmp_str = f'# Gewichtsreziprokenprobe nach Ansermet (difference = {tmp_diff}) => Passed!\n'
                if verbose:
                    print(tmp_str)
                if self.write_log:
                    self.log_str += tmp_str

        # Condition of normal equation matrix:
        if verbose or self.write_log:
            cond = np.linalg.cond(mat_N)
            tmp_str = f'Condition of normal equation matix N = {cond:1.3f}\n'
            if verbose:
                print(tmp_str)
            if self.write_log:
                self.log_str += tmp_str

        # A posteriori variance of unit weight s02:
        dof = mat_A.shape[0] - mat_A.shape[1]  # degree of freedom
        par_r = mat_v.T @ mat_P @ mat_v  # = v^T * P * v
        if dof == 0:
            # s02_a_posteriori_mugal2 = par_r[0][0]
            raise AssertionError('Degree of freedom has to be larger than 0!')
        else:
            s02_a_posteriori_mugal2 = par_r[0][0] / dof  # Eq. (20)

        s0_mugal = np.sqrt(s02_a_posteriori_mugal2)  # A posteriori std. deviation of unit weight
        if verbose or self.write_log:
            tmp_str = f'\n'
            tmp_str += f'A posteriori variance (sd) of unit weight: ' \
                       f'{s02_a_posteriori_mugal2:5.3f} ({s0_mugal:5.3f})\n'
            if verbose:
                print(tmp_str)
            if self.write_log:
                self.log_str += tmp_str

        # ### Statistics and tests ###

        # Convert co-factor matrices to covariance matrices (variances in diagonal vector)
        mat_Cvv = s02_a_posteriori_mugal2 * mat_Qvv  # A posteriori Covariance matrix of post-fit residuals
        mat_Cxx = s02_a_posteriori_mugal2 * mat_Qxx  # A posteriori Covariance matrix of estimated paramaters
        mat_Cldld = s02_a_posteriori_mugal2 * mat_Qldld  # A posteriori Covariance matrix of adjusted observations
        # diag_Qxx = np.diag(mat_Qxx)

        # Calculate standard deviations:
        mat_sd_xx = np.sqrt(np.diag(mat_Cxx))  # A posteriori SD of estimates
        mat_sd_ldld = np.sqrt(np.diag(mat_Cldld))  # A posteriori SD of adjusted observations
        # mat_sd_vv = np.sqrt(np.diag(mat_Cvv))  # A posteriori SD of residuals
        # mat_sd_vv = np.sqrt(np.diag(abs(mat_Cvv)))  # !!!! without "abs()" the sqrt operation fails because neg. values may occur!

        # A posteriori SD of residuals:
        # - Check just in case, whether all diagonal elements of the Qvv matrix are positive!
        if (np.diag(mat_Qvv) < 0).any():
            mat_sd_vv = np.sqrt(np.diag(abs(mat_Cvv)))
            if verbose or self.write_log:
                tmp_str = f' - Warning: At least one diagonal element of the Qvv matrix is negative!\n'
                if verbose:
                    print(tmp_str)
                if self.write_log:
                    self.log_str += tmp_str
        else:
            mat_sd_vv = np.sqrt(np.diag(mat_Cvv))

        # creating histogram from residuals
        residual_hist, bin_edges = create_hist(mat_v)  # Calculate histogram

        # goodness-of-fit test
        chi_crit, chi_val, chi_test = global_model_test(confidence_level_chi_test, dof,
                                                        s02_a_posteriori_mugal2, sig0_mugal ** 2)
        if verbose or self.write_log:
            tmp_str = f'\n'
            tmp_str += f'# Goodness-of-fit test results:\n'
            tmp_str += f'# Chi-val  Chi-crt-lower  Chi-crt-upper  Status\n'
            tmp_str += f'  {chi_val:7.3f}  {chi_crit[0]:13.3f}  {chi_crit[1]:13.3f}  {chi_test:s}\n'
            tmp_str += f'\n'
            tmp_str += f'# Histogram of the residuals:\n'
            tmp_str += f'# Lower-edge(µGal)  Upper-edge(µGal)  Frequency\n'
            for loop_1 in range(len(residual_hist)):
                tmp_str += f'  {bin_edges[loop_1]:16.4f}  {bin_edges[loop_1 + 1]:16.4f}  {residual_hist[loop_1]:9.0f}\n'
            if verbose:
                print(tmp_str)
            if self.write_log:
                self.log_str += tmp_str

        # Redundanzanteile (redundancy components):
        # - Measure for the outlier detection effectiveness
        # - AG II, pp. 66-71
        mat_r = np.diag(mat_Qvv @ mat_P)

        # Standardized residuals used as test statistics for outlier detection:
        # - AG II, p. 66
        # - Taking care of zero-elements in the mat_sd_vv vector to prevent division by zero errors!
        mat_w = np.zeros(len(mat_v))
        tmp_filter = ~np.isclose(mat_v[:, 0], 0.0)
        mat_w[tmp_filter] = mat_v[tmp_filter, 0] / mat_sd_vv[tmp_filter]

        # Tau test for outlier detection:
        alpha_tau = 1 - confidence_level_tau_test
        tau_test_result, tau_critical_value = tau_test(mat_w=mat_w, dof=dof, alpha=alpha_tau, mat_r=mat_r)
        number_of_outliers = tau_test_result.count("failed")

        if verbose or self.write_log:
            tmp_str = f'\n'
            tmp_str += f'# Tau-test results:\n'
            tmp_str += f'Critical value (for testing w): {tau_critical_value:1.3f}\n'
            tmp_str += f' - Number of detected outliers: {number_of_outliers}\n'
            tmp_str += f' - Number low redundancy component: {tau_test_result.count("r too small")}\n'
            tmp_str += f'\n'
            if verbose:
                print(tmp_str)
            if self.write_log:
                self.log_str += tmp_str

        # #### Store results ####
        drift_pol_coeff = mat_x[:drift_pol_degree + 1, 0]
        drift_pol_coeff_sd = mat_sd_xx[:drift_pol_degree + 1]
        vg_pol_coeff = mat_x[drift_pol_degree + 1:, 0]
        vg_pol_coeff_sd = mat_sd_xx[drift_pol_degree + 1:]
        sd_g_obs_est_mugal = mat_sd_ldld
        v_obs_est_mugal = mat_v
        sd_v_obs_est_mugal = mat_sd_vv
        w_obs_est_mugal = mat_w
        r_obs_est = mat_r
        tau_test_result_obs = tau_test_result

        # Drift parameters:
        survey_name_list = []
        degree_list = []
        coefficient_list = []
        sd_coeff_list = []
        coeff_unit_list = []
        ref_epoch_t0_dt_list = []
        tmp_idx = 0
        x_estimate_drift_coeff_names = []
        for survey_name, setup_data in self.setups.items():
            for degree in range(drift_pol_degree + 1):
                survey_name_list.append(survey_name)
                degree_list.append(degree)  # starts with 1
                coefficient_list.append(drift_pol_coeff[tmp_idx])  # * (3600**(degree + 1))  # [µGal/h]
                sd_coeff_list.append(drift_pol_coeff_sd[tmp_idx])
                ref_epoch_t0_dt_list.append(setup_data['ref_epoch_delta_t_h'])
                if degree == 0:
                    coeff_unit_list.append(f'µGal')
                else:
                    coeff_unit_list.append(f'µGal/h^{degree}')
                tmp_idx += 1
                x_estimate_drift_coeff_names.append(f'drift-{degree}')
        self.drift_pol_df = pd.DataFrame(list(zip(survey_name_list,
                                                  degree_list,
                                                  coefficient_list,
                                                  sd_coeff_list,
                                                  coeff_unit_list,
                                                  ref_epoch_t0_dt_list)),
                                         columns=self._DRIFT_POL_DF_COLUMNS)

        # VG parameters:
        degree_list = []
        coefficient_list = []
        sd_coeff_list = []
        coeff_unit_list = []
        tmp_idx = 0
        x_estimate_vg_coeff_names = []
        for degree in range(1, vg_polynomial_degree + 1):
            degree_list.append(degree)  # starts with 1
            coefficient_list.append(vg_pol_coeff[tmp_idx])  # * (3600**(degree + 1))  # [µGal/h]
            sd_coeff_list.append(vg_pol_coeff_sd[tmp_idx])
            if degree == 0:
                coeff_unit_list.append(f'µGal')
            else:
                coeff_unit_list.append(f'µGal/m^{degree}')
            tmp_idx += 1
            x_estimate_vg_coeff_names.append(f'vg-{degree}')
        self.vg_pol_df = pd.DataFrame(list(zip(degree_list,
                                               coefficient_list,
                                               sd_coeff_list,
                                               coeff_unit_list)),
                                      columns=self._VG_POL_DF_COLUMNS)

        # Observation-related results:
        for idx, v_obs_mugal in enumerate(v_obs_est_mugal):
            filter_tmp = self.setup_obs_df['obs_id'] == idx
            self.setup_obs_df.loc[filter_tmp, 'v_obs_est_mugal'] = v_obs_mugal
            self.setup_obs_df.loc[filter_tmp, 'sd_v_obs_est_mugal'] = sd_v_obs_est_mugal[idx]
            self.setup_obs_df.loc[filter_tmp, 'sd_g_obs_est_mugal'] = sd_g_obs_est_mugal[idx]
            self.setup_obs_df.loc[filter_tmp, 'w_obs_est_mugal'] = w_obs_est_mugal[idx]  # standardized residuals
            self.setup_obs_df.loc[filter_tmp, 'r_obs_est'] = r_obs_est[idx]  # redundancy components
            self.setup_obs_df.loc[filter_tmp, 'tau_test_result'] = tau_test_result_obs[idx]  # str

        # Print results to terminal and to log string:
        if verbose or self.write_log:
            tmp_str = f'\n'
            tmp_str += f' - Station data:\n'
            tmp_str += self.stat_obs_df[['station_name', 'dhf_sensor_mean_m', 'dhf_sensor_std_m',
                                         'dhf_sensor_min_m', 'dhf_sensor_max_m']].to_string(index=False,
                                                                      float_format=lambda x: '{:.4f}'.format(x))
            tmp_str += f'\n\n'
            tmp_str += f' - Drift polynomial coefficients:\n'
            tmp_str += self.drift_pol_df.to_string(index=False, float_format=lambda x: '{:.6f}'.format(x))
            tmp_str += f'\n\n'
            tmp_str += f' - VG polynomial coefficients:\n'
            tmp_str += self.vg_pol_df.to_string(index=False, float_format=lambda x: '{:.6f}'.format(x))
            tmp_str += f'\n\n'
            tmp_str += f' - Observations:\n'
            for survey_name in survey_names:
                filter_tmp = self.setup_obs_df['survey_name'] == survey_name
                tmp_str += f'   - Survey: {survey_name}\n'
            tmp_str += self.setup_obs_df.loc[filter_tmp, ['station_name', 'g_obs_mugal',
                                                          'sd_g_obs_mugal', 'sd_g_obs_est_mugal',
                                                          'v_obs_est_mugal']].to_string(index=False,
                                                                                        float_format=lambda
                                                                                            x: '{:.1f}'.format(x))
            if verbose:
                print(tmp_str)
            if self.write_log:
                self.log_str += tmp_str

        # Save data/infos to object for later use:
        self.drift_polynomial_degree = drift_pol_degree
        self.sig0_a_priori = sig0_mugal
        self.confidence_level_chi_test = confidence_level_chi_test
        self.confidence_level_tau_test = confidence_level_chi_test
        self.number_of_stations = number_of_stations
        self.number_of_estimates = number_of_parameters
        self.degree_of_freedom = dof
        self.s02_a_posteriori = s02_a_posteriori_mugal2
        self.Cxx = mat_Cxx
        self.x_estimate_names = x_estimate_drift_coeff_names + x_estimate_vg_coeff_names
        self.global_model_test_status = chi_test
        self.number_of_outliers = number_of_outliers
        self.drift_ref_epoch_type = 'survey'  # Not relevant anyway, because only ONE survey allowed in the campaign!
        self.vg_polynomial_ref_height_offset_m = vg_polynomial_ref_height_offset_m
        self.vg_polynomial_degree = vg_polynomial_degree
        self.mat_A = mat_A
        self.mat_x = mat_x
        # The following attributes are None as initialized:
        # self.scaling_factor_datum_observations
        # self.number_of_datum_stations



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'gravtools/gui/dialog_options.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Dialog_options(object):
    def setupUi(self, Dialog_options):
        Dialog_options.setObjectName("Dialog_options")
        Dialog_options.resize(319, 192)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(Dialog_options)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.groupBox_gui_options = QtWidgets.QGroupBox(Dialog_options)
        self.groupBox_gui_options.setObjectName("groupBox_gui_options")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.groupBox_gui_options)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setObjectName("formLayout")
        self.label = QtWidgets.QLabel(self.groupBox_gui_options)
        self.label.setObjectName("label")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label)
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.radioButton_gui_mode_simple = QtWidgets.QRadioButton(self.groupBox_gui_options)
        self.radioButton_gui_mode_simple.setChecked(True)
        self.radioButton_gui_mode_simple.setObjectName("radioButton_gui_mode_simple")
        self.buttonGroup_gui_mode = QtWidgets.QButtonGroup(Dialog_options)
        self.buttonGroup_gui_mode.setObjectName("buttonGroup_gui_mode")
        self.buttonGroup_gui_mode.addButton(self.radioButton_gui_mode_simple)
        self.verticalLayout.addWidget(self.radioButton_gui_mode_simple)
        self.radioButton_gui_mode_advanced = QtWidgets.QRadioButton(self.groupBox_gui_options)
        self.radioButton_gui_mode_advanced.setObjectName("radioButton_gui_mode_advanced")
        self.buttonGroup_gui_mode.addButton(self.radioButton_gui_mode_advanced)
        self.verticalLayout.addWidget(self.radioButton_gui_mode_advanced)
        self.formLayout.setLayout(0, QtWidgets.QFormLayout.FieldRole, self.verticalLayout)
        self.verticalLayout_2.addLayout(self.formLayout)
        self.verticalLayout_3.addWidget(self.groupBox_gui_options)
        self.buttonBox = QtWidgets.QDialogButtonBox(Dialog_options)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout_3.addWidget(self.buttonBox)

        self.retranslateUi(Dialog_options)
        self.buttonBox.accepted.connect(Dialog_options.accept) # type: ignore
        self.buttonBox.rejected.connect(Dialog_options.reject) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(Dialog_options)

    def retranslateUi(self, Dialog_options):
        _translate = QtCore.QCoreApplication.translate
        Dialog_options.setWindowTitle(_translate("Dialog_options", "Options"))
        self.groupBox_gui_options.setTitle(_translate("Dialog_options", "Graphical user interface (GUI) options"))
        self.label.setToolTip(_translate("Dialog_options", "Select the level of details that is preented in the GUI. The advanced mode provides more options."))
        self.label.setText(_translate("Dialog_options", "GUI mode"))
        self.radioButton_gui_mode_simple.setToolTip(_translate("Dialog_options", "Simple GUI appearance."))
        self.radioButton_gui_mode_simple.setText(_translate("Dialog_options", "Simple mode"))
        self.radioButton_gui_mode_advanced.setToolTip(_translate("Dialog_options", "Advanced GUI appearance."))
        self.radioButton_gui_mode_advanced.setText(_translate("Dialog_options", "Advanced mode"))



"""
gravtools
=========

Code by Andreas Hellerschmied
andeas.hellerschmid@bev.gv.at

Summary
-------
Command line interface o gravtools.

This module contains functions that are accessible from the command line via console_scripts entry points of setuptool.

References
----------
.. https://python-packaging.readthedocs.io/en/latest/command-line-scripts.html
"""


from bev_legacy.schwaus import main as schwaus_main
import argparse
import os
from bev_legacy.settings import OUT_PATH, PATH_OESGN_TABLE, NAME_OESGN_TABLE


def is_file(filename):
    """Check, whether the input string is the path to an existing file."""
    if os.path.isfile(filename):
        return filename
    raise argparse.ArgumentTypeError("'{}' is not a valid file.".format(filename))


def is_dir(pathname):
    """Check, whether the input string is a valid and existing filepath."""
    if os.path.exists(pathname):
        return pathname
    raise argparse.ArgumentTypeError("'{}' is not a valid directory.".format(pathname))


def schwaus():
    """Command line interface including argument parser for the legacy BEV gravity survey processing scheme."""
    parser = argparse.ArgumentParser(prog="schwaus",
                                     description="Legacy gravity survey processing scheme at BEV.",
                                     epilog="The observatios are drift-corrected by fitting a polynomial of "
                                            "degree 1 to 3 using multiple linear regression. In the second step, the "
                                            "corrected observations are fitted onto the ÖSGN level.",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--obs-file", type=is_file, required=True, help="path and name of observation file")
    parser.add_argument("--oesgn-table", required=False, type=is_file, help="path and name of ÖSGN table",
                        default=PATH_OESGN_TABLE + NAME_OESGN_TABLE)
    parser.add_argument("--out-dir", type=is_dir, required=False,
                        help="path to output directory", default=OUT_PATH)
    args = parser.parse_args()
    # print(args)

    # Prepare paths and file names for function call:
    path_oesgn_table_str, name_oesgn_table_str = os.path.split(args.oesgn_table)
    path_obs_file_str, name_obs_file_str = os.path.split(args.obs_file)
    if path_oesgn_table_str:  # String not empty => Append '/'
        path_oesgn_table_str += '/'
    if path_obs_file_str:  # String not empty => Append '/'
        path_obs_file_str += '/'
    args.out_dir += '/'

    return schwaus_main(path_oesgn_table_str, name_oesgn_table_str, path_obs_file_str, name_obs_file_str, args.out_dir)




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'gravtools/gui/dialog_gis_export_settings.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Dialog_gis_settings(object):
    def setupUi(self, Dialog_gis_settings):
        Dialog_gis_settings.setObjectName("Dialog_gis_settings")
        Dialog_gis_settings.resize(674, 342)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(Dialog_gis_settings)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.groupBox_general_settings = QtWidgets.QGroupBox(Dialog_gis_settings)
        self.groupBox_general_settings.setObjectName("groupBox_general_settings")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.groupBox_general_settings)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.formLayout_2 = QtWidgets.QFormLayout()
        self.formLayout_2.setObjectName("formLayout_2")
        self.label = QtWidgets.QLabel(self.groupBox_general_settings)
        self.label.setToolTip("EPSG code that defines the coordiante reference system of the station coordinates loaded from the station data file or from the observation input file(s).")
        self.label.setObjectName("label")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label)
        self.lineEdit_stat_coord_epsg = QtWidgets.QLineEdit(self.groupBox_general_settings)
        self.lineEdit_stat_coord_epsg.setObjectName("lineEdit_stat_coord_epsg")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.lineEdit_stat_coord_epsg)
        self.verticalLayout_2.addLayout(self.formLayout_2)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_2.addItem(spacerItem)
        self.verticalLayout_3.addWidget(self.groupBox_general_settings)
        self.groupBox_export_lsm_run_results = QtWidgets.QGroupBox(Dialog_gis_settings)
        self.groupBox_export_lsm_run_results.setObjectName("groupBox_export_lsm_run_results")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.groupBox_export_lsm_run_results)
        self.verticalLayout.setObjectName("verticalLayout")
        self.formLayout_3 = QtWidgets.QFormLayout()
        self.formLayout_3.setObjectName("formLayout_3")
        self.radioButton_campaign_output_dir = QtWidgets.QRadioButton(self.groupBox_export_lsm_run_results)
        self.radioButton_campaign_output_dir.setChecked(True)
        self.radioButton_campaign_output_dir.setObjectName("radioButton_campaign_output_dir")
        self.buttonGroup_output_dir = QtWidgets.QButtonGroup(Dialog_gis_settings)
        self.buttonGroup_output_dir.setObjectName("buttonGroup_output_dir")
        self.buttonGroup_output_dir.addButton(self.radioButton_campaign_output_dir)
        self.formLayout_3.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.radioButton_campaign_output_dir)
        self.lineEdit_output_subdir = QtWidgets.QLineEdit(self.groupBox_export_lsm_run_results)
        self.lineEdit_output_subdir.setObjectName("lineEdit_output_subdir")
        self.formLayout_3.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.lineEdit_output_subdir)
        self.radioButton_selected_output_dir = QtWidgets.QRadioButton(self.groupBox_export_lsm_run_results)
        self.radioButton_selected_output_dir.setObjectName("radioButton_selected_output_dir")
        self.buttonGroup_output_dir.addButton(self.radioButton_selected_output_dir)
        self.formLayout_3.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.radioButton_selected_output_dir)
        self.verticalLayout.addLayout(self.formLayout_3)
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setObjectName("formLayout")
        self.pushButton_select_gis_output_dir = QtWidgets.QPushButton(self.groupBox_export_lsm_run_results)
        self.pushButton_select_gis_output_dir.setEnabled(False)
        self.pushButton_select_gis_output_dir.setObjectName("pushButton_select_gis_output_dir")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.pushButton_select_gis_output_dir)
        self.lineEdit_gis_output_dir = QtWidgets.QLineEdit(self.groupBox_export_lsm_run_results)
        self.lineEdit_gis_output_dir.setEnabled(False)
        self.lineEdit_gis_output_dir.setObjectName("lineEdit_gis_output_dir")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.lineEdit_gis_output_dir)
        self.checkBox_export_stat_results_shp = QtWidgets.QCheckBox(self.groupBox_export_lsm_run_results)
        self.checkBox_export_stat_results_shp.setChecked(True)
        self.checkBox_export_stat_results_shp.setObjectName("checkBox_export_stat_results_shp")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.checkBox_export_stat_results_shp)
        self.lineEdit_filename_stat_results_shp = QtWidgets.QLineEdit(self.groupBox_export_lsm_run_results)
        self.lineEdit_filename_stat_results_shp.setObjectName("lineEdit_filename_stat_results_shp")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.lineEdit_filename_stat_results_shp)
        self.checkBox_export_obs_results_shp = QtWidgets.QCheckBox(self.groupBox_export_lsm_run_results)
        self.checkBox_export_obs_results_shp.setChecked(True)
        self.checkBox_export_obs_results_shp.setObjectName("checkBox_export_obs_results_shp")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.checkBox_export_obs_results_shp)
        self.lineEdit_filename_obs_results_shp = QtWidgets.QLineEdit(self.groupBox_export_lsm_run_results)
        self.lineEdit_filename_obs_results_shp.setObjectName("lineEdit_filename_obs_results_shp")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.lineEdit_filename_obs_results_shp)
        self.checkBox_add_lsm_method_filename = QtWidgets.QCheckBox(self.groupBox_export_lsm_run_results)
        self.checkBox_add_lsm_method_filename.setChecked(True)
        self.checkBox_add_lsm_method_filename.setObjectName("checkBox_add_lsm_method_filename")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.checkBox_add_lsm_method_filename)
        self.verticalLayout.addLayout(self.formLayout)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem1)
        spacerItem2 = QtWidgets.QSpacerItem(20, 6, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem2)
        self.verticalLayout_3.addWidget(self.groupBox_export_lsm_run_results)

        self.retranslateUi(Dialog_gis_settings)
        self.radioButton_selected_output_dir.toggled['bool'].connect(self.pushButton_select_gis_output_dir.setEnabled) # type: ignore
        self.radioButton_selected_output_dir.toggled['bool'].connect(self.lineEdit_gis_output_dir.setEnabled) # type: ignore
        self.checkBox_export_obs_results_shp.toggled['bool'].connect(self.lineEdit_filename_obs_results_shp.setEnabled) # type: ignore
        self.checkBox_export_stat_results_shp.toggled['bool'].connect(self.lineEdit_filename_stat_results_shp.setEnabled) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(Dialog_gis_settings)

    def retranslateUi(self, Dialog_gis_settings):
        _translate = QtCore.QCoreApplication.translate
        Dialog_gis_settings.setWindowTitle(_translate("Dialog_gis_settings", "GIS data export settings"))
        self.groupBox_general_settings.setTitle(_translate("Dialog_gis_settings", "General settings"))
        self.label.setText(_translate("Dialog_gis_settings", "EPSG code of station coordinates"))
        self.lineEdit_stat_coord_epsg.setToolTip(_translate("Dialog_gis_settings", "EPSG code that defines the coordiante reference system of the station coordinates loaded from the station data file or from the observation input file(s)."))
        self.groupBox_export_lsm_run_results.setToolTip(_translate("Dialog_gis_settings", "Settings for the GIS data export in the Results tab (export of LSM run results for data visualization)."))
        self.groupBox_export_lsm_run_results.setTitle(_translate("Dialog_gis_settings", "Export for visualizing LSM run results"))
        self.radioButton_campaign_output_dir.setText(_translate("Dialog_gis_settings", "Write GIS files to campaign output directory"))
        self.lineEdit_output_subdir.setToolTip(_translate("Dialog_gis_settings", "<html><head/><body><p>Subdirectory in the campaign\'s output directory for writing the results shapefiles.</p></body></html>"))
        self.radioButton_selected_output_dir.setText(_translate("Dialog_gis_settings", "Write GIS files to different output directroy"))
        self.pushButton_select_gis_output_dir.setToolTip(_translate("Dialog_gis_settings", "Select export directory"))
        self.pushButton_select_gis_output_dir.setText(_translate("Dialog_gis_settings", "Select export directory"))
        self.lineEdit_gis_output_dir.setToolTip(_translate("Dialog_gis_settings", "Export directory (derfault: campaign output directory)"))
        self.checkBox_export_stat_results_shp.setText(_translate("Dialog_gis_settings", "Export station results (shapefile)"))
        self.lineEdit_filename_stat_results_shp.setToolTip(_translate("Dialog_gis_settings", "Filename without \".shp\""))
        self.checkBox_export_obs_results_shp.setText(_translate("Dialog_gis_settings", "Export observation results (shapefile)"))
        self.lineEdit_filename_obs_results_shp.setToolTip(_translate("Dialog_gis_settings", "Filename without \".shp\""))
        self.checkBox_add_lsm_method_filename.setToolTip(_translate("Dialog_gis_settings", "Add the LSM Method to the fileanem string. This is need e.g. to distinguished shapefiles created by from LSM runs using different estimation methods."))
        self.checkBox_add_lsm_method_filename.setText(_translate("Dialog_gis_settings", "Add LSM Method to filename"))



"""Misc. methods for the GUI.

Copyright (C) 2021  Andreas Hellerschmied <andreas.hellerschmied@bev.gv.at>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import pyqtgraph as pg
from PyQt5.QtCore import Qt
import random


def get_station_color_dict(stations: list, randomize=False) -> dict:
    """Returns a dict with unique pyqtgraph QColor items for each station in the input list."""
    station_color_dict = {}
    number_of_stations = len(stations)

    stat_id_list = [i for i in range(0, number_of_stations)]
    if randomize:
        random.seed(1)  # Use always the same seed to get reproducibility
        random.shuffle(stat_id_list)

    for stat_id, station in enumerate(stations, start=0):
        stat_id_altered = stat_id_list[stat_id]
        station_color_dict[station] = pg.intColor(stat_id_altered, hues=number_of_stations)
    return station_color_dict


def checked_state_to_bool(checked_state) -> bool:
    """Converts Qt checked states to boolean values."""
    if checked_state == Qt.Checked or checked_state == Qt.PartiallyChecked:
        return True
    elif checked_state == Qt.Unchecked:
        return False
    else:
        raise AttributeError('Invalid input argument!')


def resize_table_view_columns(table_view, n, add_pixel=0):
    """Resize the width of columns of a QTableView based on the header and the content of the first n rows.

    Parameters
    ----------
    table_view: QTableView object
        The QTableView object that should be concerned.
    n : int
        The column width is adjusted to the width of the column headers and the content of the first n rows. Is the
        model contains less than n items, all available items are considered.
    add_pixel : int >= 0
        Number of pixels to be added to the minimum width to improve readability.
    """
    fm = table_view.fontMetrics()
    for col_num in range(0, table_view.model().columnCount()):
        name= table_view.model().headerData(section=col_num, orientation=Qt.Horizontal, role=Qt.DisplayRole)
        max_width = fm.width(name)
        for row_idx in range(0, n):
            data_index = table_view.model().index(row_idx, col_num)
            if not data_index.isValid():
                break
            data = table_view.model().data(data_index)
            width_data = fm.width(data)
            if width_data > max_width:
                max_width = width_data
        table_view.setColumnWidth(col_num, max_width + add_pixel)


if __name__ == "__main__":
    """Main Program for debugging."""
    stat_list = ['stat1', 'stat2', 'stat3', 'stat4']
    station_color_dict = get_station_color_dict(stat_list)
    print(station_color_dict)






"""
gravtools
=========

Code by Andreas Hellerschmied
andeas.hellerschmid@bev.gv.at

Summary
-------
User-defined settings for gravtools.
"""

# Additive constant for the determination of the full absolute gravity [µGal] from observed values:
ADDITIVE_CONST_ABS_GRAVITY = 9.8e8

SURVEY_DATA_SOURCE_TYPES = {
    'cg5_obs_file_txt': 'Scintrex CG5 observation file (text format)',
    'bev_obs_file': 'Simple observation file format used by BEV',
}

STATION_DATA_SOURCE_TYPES = {
    'oesgn_table': 'Control points of the Austrian gravity base network (OESGN).',
    'obs_file': 'From an observation file'
}

TIDE_CORRECTION_TYPES = {
    'instrumental_corr': 'Instrument-implemented tidal correction',
    'no_tide_corr': 'No tide correction applied',
    'unknown': 'Unknown whether a tide correction was applied',
}

REFERENCE_HEIGHT_TYPE = {
    'sensor_height': 'The gravity reverence point at the station is at the sensor height',
    'instrument_top': 'The gravity value refers to the height og the instrument top',
    'ground': 'The gravity value refers to the ground point at the station',
    'control_point': 'The gravity value refers to the control point at the station',
}

# Reference heights (sensor heights) of different gravimeter types:
GRAVIMETER_REFERENCE_HEIGHT_CORRECTIONS_m = {
    'CG3': -0.211,
    'CG5': -0.211,
}

# Valid gravimeter types and description.
GRAVIMETER_TYPES = {
    'CG5': 'Sctintrex CG5',
    'CG3': 'Sctintrex CG3',
}

# Default Gavimeter Type when loading data from an CG5 observation file
DEFAULT_GRAVIMETER_TYPE_CG5_SURVEY = 'CG5'

# Lookuptable to convert gravimeter type to Kennzeichen used at BEV (in the database NSDB):
GRAVIMETER_TYPES_KZG_LOOKUPTABLE = {
    'CG5': 'C',
    'CG3': 'C',
}

# Valid Gravimeter serial numbers (S/N and type)
GRAVIMETER_SERIAL_NUMBERS = {
    '40601': 'CG5'
}

# Lookuptable to convert the gravimeter S/N to the IDs written to the database NSDB:
GRAVIMETER_SERIAL_NUMBER_TO_ID_LOOKUPTABLE = {
    '40601': '5'
}


#--------------

# Instrumenten-IDs in der Messdatei einem Instrument zuweisen:
GRAVIMETER_ID_BEV = {
    '3': 'CG3',
    '5': 'CG5',
}

# Instrumenten-IDs in der Messdatei einem Instrumenten-KZ zuweisen:
GRAVIMETER_KZ_BEV = {
    '3': 'C',
    '5': 'C',
    '9': 'D',
    '90': 'D',
    '900': 'D',
    '51': 'D',
    '510': 'D',
    '500': 'W',
}

#--------------------

#  Lookup table for matching gravimeter IDs and the tidal corrections that are applied per default in the BEV legacy
#  observation files:
BEV_GRAVIMETER_TIDE_CORR_LOOKUP = {
    '5': 'instrumental_corr'
}

# Available adjustment methods
ADJUSTMENT_METHODS = {
    'LSM_diff': 'LSM (differential observations)',  # Least-squares adjustment of differential observations
    'MLR_BEV': 'MLR (BEV legacy processing)',  # Least-squares adjustment of differential observations
}

# Treshold for the "Gewichtsreziprokenprobe nach Ansermet" (see Skriptum AG1, p. 136, Eq. (6.86))
ANSERMET_DIFF_THRESHOLD = 1e-3

# Treshold for the redundancy component of an observation in order to apply a pope test for outlier detection:
R_POPE_TEST_THRESHOLD = 1e-6

# Only consider active observations for the determination of the reference epochs, e.g. for the drift polynomial. The
# Reference epochs are determined based on the first (active only or active/inactive) observations in the campaign or
# in each individual survey, depending on the settings.
ACTIVE_OBS_ONLY_FOR_REF_EPOCH = True

# GUI and Program options:
CALCULATE_REDUCED_OBS_WHEN_LOADING_DATA = True  # Calculate reduced observations when loading observation data.

# SCHWAUS and DRIFT settings:
FLAG_SAVE_DRIFT_PLOT_PDF = True
FLAG_CREATE_SCHWAUS_PROTOCOL = True
VERBOSE = True

# Output directory
OUT_PATH = '/home/heller/pyProjects/GravTools/out/'

# Default names and paths of input files:

# ÖSGN Table:
PATH_OESGN_TABLE = '/home/heller/pyProjects/GravTools/data/'
NAME_OESGN_TABLE = 'OESGN_f200701.tab'

# BEV observation files:
PATH_OBS_FILE_BEV = '/home/heller/pyProjects/GravTools/data/BEV/'
# NAME_OBS_FILE_BEV = '20200527_tideCorr'
# NAME_OBS_FILE_BEV = '20200527_sd'
# NAME_OBS_FILE_BEV = '20200527_2'
# NAME_OBS_FILE_BEV = '20200527'
#NAME_OBS_FILE_BEV = 'n20200701_1'
#NAME_OBS_FILE_BEV = 'f200701_1'
NAME_OBS_FILE_BEV = 'f200701_3'
# NAME_OBS_FILE_BEV = 'e201001'

# CG-5 observation files (text)
PATH_OBS_FILE_CG5 = '/home/heller/pyProjects/GravTools/data/CG5/'
# NAME_OBS_FILE_CG5 = '2020-06-18_DACH.TXT'
NAME_OBS_FILE_CG5 = '20200907_test.TXT'
